from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from six import text_type

from aspen.http.request import Path, Querystring


# Path
# ====

def test_path_starts_empty():
    path = Path("/bar.html")
    assert path == {}, path

def test_path_has_raw_set():
    path = Path("/bar.html")
    assert path.raw == "/bar.html", path.raw

def test_path_raw_is_unicode():
    path = Path("/bar.html")
    assert isinstance(path.raw, text_type)

def test_path_has_decoded_set():
    path = Path("/b%c3%a4r.html")
    assert path.decoded == "/b\u00e4r.html", path.decoded

def test_path_decoded_is_unicode():
    path = Path("/bar.html")
    assert isinstance(path.decoded, text_type)

def test_path_unquotes_and_decodes_UTF_8():
    path = Path("/%e2%98%84.html")
    assert path.decoded == "/\u2604.html", path.decoded

def test_path_doesnt_unquote_plus():
    path = Path("/+%2B.html")
    assert path.decoded == "/++.html", path.decoded

def test_path_has_parts():
    path = Path("/foo/bar.html")
    assert path.parts == ['foo', 'bar.html']


# Path params
# ===========

def _extract_params(path):
#    return dispatcher.extract_rfc2396_params(path.lstrip('/').split('/'))
    params = [ segment.params for segment in path.parts ]
    segments = [ str(segment) for segment in path.parts ]
    return ( segments, params )

def test_extract_path_params_with_none():
    path = Path("/foo/bar")
    actual = _extract_params(path)
    expected = (['foo', 'bar'], [{}, {}])
    assert actual == expected

def test_extract_path_params_simple():
    path = Path("/foo;a=1;b=2;c/bar;a=2;b=1")
    actual = _extract_params(path)
    expected = (['foo', 'bar'], [{'a':['1'], 'b':['2'], 'c':['']}, {'a':['2'], 'b':['1']}])
    assert actual == expected

def test_extract_path_params_complex():
    path = Path("/foo;a=1;b=2,3;c;a=2;b=4/bar;a=2,ab;b=1")
    actual = _extract_params(path)
    expected = (['foo', 'bar'], [{'a':['1','2'], 'b':['2,3', '4'], 'c':['']}, {'a':[ '2,ab' ], 'b':['1']}])
    assert actual == expected

def test_path_params_api():
    path = Path("/foo;a=1;b=2;b=3;c/bar;a=2,ab;b=1")
    parts, params = (['foo', 'bar'], [{'a':['1'], 'b':['2', '3'], 'c':['']}, {'a':[ '2,ab' ], 'b':['1']}])
    assert path.parts == parts, path.parts
    assert path.parts[0].params == params[0]
    assert path.parts[1].params == params[1]


# Querystring
# ===========

def test_querystring_starts_full():
    querystring = Querystring("baz=buz")
    assert querystring == {'baz': ['buz']}, querystring

def test_querystring_has_raw_set():
    querystring = Querystring("baz=buz")
    assert querystring.raw == "baz=buz", querystring.raw

def test_querystring_raw_is_unicode():
    querystring = Querystring("baz=buz")
    assert isinstance(querystring.raw, text_type)

def test_querystring_has_decoded_set():
    querystring = Querystring("baz=buz")
    assert querystring.decoded == "baz=buz", querystring.decoded

def test_querystring_decoded_is_unicode():
    querystring = Querystring("baz=buz")
    assert isinstance(querystring.decoded, text_type)

def test_querystring_unquotes_and_decodes_UTF_8():
    querystring = Querystring("baz=%e2%98%84")
    assert querystring.decoded == "baz=\u2604", querystring.decoded

def test_querystring_comes_out_UTF_8():
    querystring = Querystring("baz=%e2%98%84")
    assert querystring['baz'] == "\u2604", querystring['baz']

def test_querystring_doesnt_choke_on_bad_unicode():
    querystring = Querystring("baz%e2%98=%e2%98")
    assert querystring['baz\ufffd'] == '\ufffd'

def test_querystring_unquotes_plus():
    querystring = Querystring("baz=+%2B")
    assert querystring.decoded == "baz= +", querystring.decoded
    assert querystring['baz'] == " +"



"""Wherein our hero learns about sys.path_importer_cache.

Python caches some of its import machinery, and if you try really hard, you can
shoot yourself in the foot with that.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import sys
from pprint import pformat


FIX_VIA_IMPORT = 1
FIX_VIA_PROJ = 0
FSFIX = os.path.realpath('fsfix')


def log(a, *b):
    return  # turn off logging now that this works
    if not b:
        print(a.rjust(30))
    else:
        for line in b:
            print(a.rjust(30)+": ", line)


def rm():
    root = FSFIX
    if os.path.isdir(root):
        for root, dirs, files in os.walk(root, topdown=False):
            for name in dirs:
                _ = os.path.join(root, name)
                log("removing  dir", _)
                os.rmdir(_)
            for name in files:
                _ = os.path.join(root, name)
                log("removing file", _)
                os.remove(_)
        log("removing root", root)
        os.rmdir(root)
    sys.path_importer_cache = {}

def __dump():
    log("sys.path_importer_cache", pformat(sys.path_importer_cache).splitlines())

def test_weirdness():
    try:
        #print
        foo = os.path.join(FSFIX, 'foo')
        foo = os.path.realpath(foo)
        if foo not in sys.path:
            log("inserting into sys.path", foo )
            sys.path.insert(0, foo)

        log("making directory", FSFIX)
        os.mkdir(FSFIX)
        if FIX_VIA_PROJ:
            log("making directory", FSFIX + '/foo')
            os.mkdir(FSFIX + '/foo')

        log("importing a thing")
        old = set(sys.path_importer_cache.keys())
        import aspen
        aspen  # shut up pyflakes
        now = set(sys.path_importer_cache.keys())
        log("diff", now - old)

        rm()

        log("making directory", FSFIX)
        os.mkdir(FSFIX)
        log("making directory", FSFIX + '/foo')
        os.mkdir(FSFIX + '/foo')
        log("making file", FSFIX + '/foo' + '/bar.py')
        open(FSFIX + '/foo/bar.py', 'w+').write('baz = "buz"')

        log("contents of fsfix/foo/bar.py", open('fsfix/foo/bar.py').read())
        log("contents of sys.path", *sys.path)

        log("importing bar")
        __dump()
        try:
            import bar
            bar  # shut up pyflakes
            log("succeeded")
        except:
            log("failed")
            raise
    finally:
        rm()

    #print

if __name__ == '__main__':
    test_weirdness()



#!/usr/bin/env python
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import sys

from six import unichr

for i in range(2**16):
    u = unichr(i).encode('utf8')
    sys.stdout.write("%5d %s  " % (i, u))
    if i % 6 == 0:
        print()




from __future__ import division, print_function, unicode_literals, with_statement

import fnmatch
import os
import shlex
import sys
import os.path
from fabricate import ExecutionError, main, run, shell, autoclean

# Core Executables
# ================

# See https://github.com/AspenWeb/pando.py/issues/542
USE_PY_VENV = sys.version_info > (3, 3)


def _virt(cmd, envdir='env'):
    envdir = _env(envdir)
    if os.name == "nt":
        return os.path.join(envdir, 'Scripts', cmd + '.exe')
    else:
        return os.path.join(envdir, 'bin', cmd)


def _virt_version(envdir):
    v = shell(_virt('python', envdir), '-c',
              'import sys; print(sys.version_info[:2])')
    return eval(v)


def _env(envdir='env'):
    d = __env(envdir)
    # extend the PATH
    path = os.path.join(d, 'Scripts' if os.name == "nt" else 'bin')
    os.environ['PATH'] = path + os.pathsep + os.environ.get('PATH', '')
    return d


def __env(envdir):
    # http://stackoverflow.com/a/1883251
    if hasattr(sys, 'real_prefix'):
        # We're already inside someone else's virtualenv.
        return sys.prefix
    elif hasattr(sys, 'base_prefix') and sys.base_prefix != sys.prefix:
        # We're already inside someone else's pyvenv.
        return sys.prefix
    elif os.path.exists(envdir):
        # We've already built our own virtualenv.
        return envdir

    if USE_PY_VENV:
        # use built-in venv module
        run(sys.executable, '-m', 'venv', envdir)
    else:
        # use virtualenv instead
        try:
            import virtualenv
        except ImportError:
            # install it when missing
            run(sys.executable, '-m', 'pip', 'install', '--user', 'virtualenv')
        run(sys.executable, '-m', 'virtualenv', envdir)
    return envdir


def env():
    """set up a base virtual environment"""
    _env()


def _install_tox():
    # install tox if it isn't there
    try:
        shell('pip', 'show', 'tox')
    except ExecutionError:
        run('pip', 'install', 'tox')


def _deps():
    shell('pip', 'install', '-r', 'requirements.txt', ignore_status=False)


def _test_deps():
    _deps()
    shell('pip', 'install', '-r', 'requirements_tests.txt', ignore_status=False)


def _dev(envdir='env'):
    envdir = _env(envdir)
    _install_tox()
    run('tox', '--notest', '--skip-missing-interpreters')
    return envdir


def dev():
    """set up an environment able to run aspen and the tests"""
    _dev()


def clean_env():
    """clean env artifacts"""
    shell('rm', '-rf', 'env')


def clean():
    """clean all artifacts"""
    autoclean()
    delete_files('*.pyc', '.')
    clean_env()
    clean_sphinx()
    clean_test()
    clean_build()


# Docs
# ====

def _sphinx_cmd(cmd, extra_args=[], extra_pkgs=[]):
    envdir = _env()
    _deps()
    run('pip', 'install', *(['sphinx', 'sphinx-rtd-theme'] + extra_pkgs))
    builddir = 'docs/_build'
    run('mkdir', '-p', builddir)
    args = ['-b', 'html', '-d', builddir + '/doctrees', 'docs', builddir + '/html']
    args += extra_args
    run(cmd, args)

def sphinx():
    """build sphinx documents"""
    _sphinx_cmd("sphinx-build")

def autosphinx():
    """run sphinx-autobuild"""
    _sphinx_cmd("sphinx-autobuild", ['--watch', 'aspen'], extra_pkgs=['sphinx-autobuild'])

def clean_sphinx():
    """clean sphinx artifacts"""
    shell('rm', '-rf', 'docs/_build')


# Testing
# =======

def _tox(*args, **kw):
    _env()
    _install_tox()
    kw.setdefault('silent', False)
    shell('tox', '--skip-missing-interpreters', '--', *args, **kw)


def test():
    """run all tests"""
    # this calls tox, and tox calls the _test target below from inside each env
    _tox(ignore_status=False)


def _test(pytest_args=()):
    _test_deps()
    delete_files('*.pyc', 'aspen', 'tests')
    pytest_args = pytest_args or shlex.split(os.environ.get('PYTEST_ARGS', ''))
    shell('python', '-m', 'pytest', 'tests', *pytest_args, ignore_status=False, silent=False)
    shell('pyflakes', 'aspen', 'tests', ignore_status=False, silent=False)


def testf():
    """run tests, stopping at the first failure"""
    _tox('python', 'build.py', '_testf', ignore_status=True)


def _testf():
    _test(pytest_args=['-x'])


def pyflakes():
    _tox('pyflakes', 'aspen', 'tests', ignore_status=False)


def pylint():
    """run lint"""
    envdir = _env()
    run(_virt('pip', envdir), 'install', 'pylint')
    run(_virt('pylint', envdir), '--rcfile=.pylintrc',
        'aspen', '|', 'tee', 'pylint.out', shell=True, ignore_status=True)


def test_cov():
    """run code coverage"""
    os.environ['PYTEST_ARGS'] = (
        '--junitxml=testresults.xml '
        '--cov-report term '
        '--cov-report xml '
        '--cov-report html '
        '--cov aspen'
    )
    test()


def analyse():
    """run lint and coverage"""
    pylint()
    test_cov()
    print('done!')


def clean_test():
    """clean test artifacts"""
    shell('rm', '-rf', '.tox')
    shell('rm', '-rf', '.coverage', 'coverage.xml', 'testresults.xml', 'htmlcov', 'pylint.out')


# Build
# =====

def build():
    """build an egg"""
    run(sys.executable, 'setup.py', 'bdist_egg')


def wheel():
    """build a wheel"""
    run(sys.executable, 'setup.py', 'bdist_wheel')


def clean_build():
    """clean build artifacts"""
    run('python', 'setup.py', 'clean', '-a')
    run('rm', '-rf', 'dist')


# Utils
# =====

def find_files(directory, pattern):
    for root, dirs, files in os.walk(directory):
        for filename in fnmatch.filter(files, pattern):
            yield os.path.join(root, filename)

def delete_files(pattern, *directories):
    for d in directories:
        for fpath in find_files(d, pattern):
            os.remove(fpath)


def show_targets():
    """show the list of valid targets (this list)"""
    print("Valid targets:\n")
    # organize these however
    targets = ['show_targets', None,
               'env', 'dev', 'testf', 'test', 'pylint', 'test_cov', 'analyse', None,
               'build', 'wheel', None,
               'sphinx', 'autosphinx', None,
               'clean', 'clean_env', 'clean_test', 'clean_build', 'clean_sphinx', None,
               ]
    #docs = '\n'.join(["  %s - %s" % (t, LOCALS[t].__doc__) for t in targets])
    #print(docs)

    for t in targets:
        if t is not None:
            print("  %s - %s" % (t, LOCALS[t].__doc__))
        else:
            print("")

    if len(targets) < (len(LOCALS) - len(NON_TARGETS)):
        missed = set(LOCALS.keys()).difference(NON_TARGETS, targets)
        print("Unordered targets: " + ', '.join(sorted(missed)))
    sys.exit()


LOCALS = dict(locals())
NON_TARGETS = [ 'main', 'autoclean', 'run', 'shell' ]
NON_TARGETS += list(x for x in LOCALS if x.startswith('_') or not callable(LOCALS[x] ))

if __name__ == '__main__':
    main( default='show_targets'
        , ignoreprefix="python"  # workaround for gh190
         )



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pytest import raises


GARBAGE = b"a\xef\xf9"


# I had thought to wrap codec error handler registration in a context manager
# so we could clean up the global registry state post-test, but:

# > There is no API to unregister a codec search function, since deregistration
# > would break the codec cache used by the registry to speedup codec
# > lookup.
#
# > Why would you want to unregister a codec search function ?
#
# https://mail.python.org/pipermail/python-dev/2011-September/113590.html

# O.o


def test_garbage_is_garbage():
    raises(UnicodeDecodeError, lambda s: s.decode('utf8'), GARBAGE)

def test_backslashreplace_error_strategy_works_when_decoding():
    actual = GARBAGE.decode('utf8', 'backslashreplace')
    assert actual == r"a\xef\xf9"

def test_backslashreplace_error_strategy_works_when_encoding():
    actual = 'comet: \u2604'.encode('ascii', 'backslashreplace')
    assert actual == br"comet: \u2604"



from __future__ import absolute_import, division, print_function, unicode_literals

import gc
import json
from timeit import timeit

from filesystem_tree import FilesystemTree

import aspen.request_processor


def is_dynamic(fspath):
    return fspath.endswith('.spt')


FILE_CONTENT = "The content doesn't matter, we're only testing dispatching."

FILES = [
    ('index.html.spt', FILE_CONTENT),
    ('style.css', FILE_CONTENT),
    ('%username/index.spt', FILE_CONTENT),
    ('foo/%catchall.spt', FILE_CONTENT),
]

URLS = [
    '/',
    '/style.css',
    '/username',
    '/username/',
    '/nonexistent/file.php',
    '/foo/',
    '/foo/bar',
    '/foo/bar.txt',
    '/foo/bar/',
    '/foo/bar/baz',
]


times = {}
for dispatcher_class in aspen.request_processor.dispatcher.DISPATCHER_CLASSES:
    print("Timing", dispatcher_class.__name__)
    total_time = 0
    with FilesystemTree() as ft:
        ft.mk(*FILES)
        dispatcher = dispatcher_class(
            ft.root,
            is_dynamic,
            aspen.request_processor.default_indices,
            aspen.request_processor.typecasting.defaults
        )
        dispatcher.build_dispatch_tree()
        for url in URLS:
            dispatch = lambda: dispatcher.dispatch(url, url.split('/'))
            time = timeit(dispatch, number=1000)
            print(url, time)
            total_time += time
    times[dispatcher_class.__name__] = total_time
    print()

print("Totals:", json.dumps(times, indent=4, sort_keys=True), end='\n\n')

ordered = sorted(times.items(), key=lambda t: t[1])
if len(ordered) > 2:
    print(
        "The winner is %s: it's %.2f times faster than the second best %s, and %.2f times faster than the slowest %s." %
        ( ordered[0][0], ordered[1][1] / ordered[0][1], ordered[1][0]
        , ordered[-1][1] / ordered[0][1], ordered[-1][0])
    )
else:
    print("The winner is %s: it's %.2f times faster than %s." %
          (ordered[0][0], ordered[1][1] / ordered[0][1], ordered[1][0]))



"""
This module handles the parsing of path variables.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from ..exceptions import TypecastError


#: Aspen's default typecasters.
defaults = { 'int': lambda pathpart, context: int(pathpart)
           , 'float': lambda pathpart, context: float(pathpart)
            }


def apply_typecasters(typecasters, path_vars, context):
    """Perform typecasting (in-place!).

    Args:
        typecasters: a :class:`dict` of type names to typecast functions
        path_vars: a :class:`~aspen.http.mapping.Mapping` of path variables
        context: a :class:`dict` passed to typecast functions as second argument

    Raises:
        TypecastError: if a typecast function raises an exception
    """
    for part in list(path_vars.keys()):
        if '.' in part:
            var, ext = part.rsplit('.', 1)
            if ext in typecasters:
                try:
                    # path_vars is a Mapping not a dict, so:
                    for v in path_vars.all(part):
                        path_vars.add(var, typecasters[ext](v, context))
                    path_vars.popall(part)
                except:
                    raise TypecastError(ext)




# -*- coding: utf-8 -*-
#
# Aspen documentation build configuration file, originally created by
# sphinx-quickstart on Mon Feb 24 15:00:43 2014.
#
# This file is exec()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this file.

import sys, os

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath('..'))


# -- General configuration -----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'sphinx.ext.autodoc', 'sphinx.ext.doctest', 'sphinx.ext.intersphinx',
    'sphinx.ext.todo', 'sphinx.ext.coverage', 'sphinx.ext.ifconfig',
    'sphinx.ext.napoleon',
]

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'Aspen'
copyright = u'2016, Chad Whitacre, et al'  # RtD theme doubles the period

# The full version, including alpha/beta/rc tags.
release = open('../version.txt').read().strip()
# The short X.Y version.
version = release.split('-', 1)[0]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'


# -- Autodoc options

autodoc_default_flags = ['members']
autodoc_member_order = 'bysource'


# -- Options for HTML output ---------------------------------------------------

import sphinx_rtd_theme
html_theme = "sphinx_rtd_theme"
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# Output file base name for HTML help builder.
htmlhelp_basename = 'Aspendoc'


# -- Options for LaTeX output --------------------------------------------------

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
  ('index', 'Aspen.tex', u'Aspen Documentation',
   u'Chad Whitacre et al.', 'manual'),
]


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'aspen', u'Aspen Documentation',
     [u'Chad Whitacre et al.'], 1)
]


# -- Options for Texinfo output ------------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
  ('index', 'Aspen', u'Aspen Documentation',
   u'Chad Whitacre et al.', 'Aspen',
   'A filesystem router for Python web frameworks.',
   'Miscellaneous'),
]


# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {'http://docs.python.org/': None}



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import io

from pytest import raises

from aspen.simplates import json_


def test_json_basically_works(harness):
    expected = '''{
    "Greetings": "program!"
}'''
    actual = harness.simple( "[---]\n[---] application/json\n{'Greetings': 'program!'}"
                           , filepath="foo.json.spt"
                            ).text
    assert actual == expected

def test_json_defaults_to_application_json_for_static_json(harness):
    actual = harness.simple( '{"Greetings": "program!"}'
                           , filepath="foo.json"
                            ).media_type
    assert actual == 'application/json'

def test_json_content_type_is_configurable_for_static_json(harness):
    harness.request_processor.media_type_json = "floober/blah"
    expected = 'floober/blah'
    actual = harness.simple( '{"Greetings": "program!"}'
                           , filepath="foo.json"
                            ).media_type
    assert actual == expected

def test_json_content_type_is_configurable_from_kwargs(harness):
    actual = harness.simple( '{"Greetings": "program!"}'
                           , filepath="foo.json"
                           , request_processor_configuration={'media_type_json': 'floober/blah'}
                            ).media_type
    assert actual == 'floober/blah'

def test_json_content_type_is_configurable_for_dynamic_json(harness):
    harness.request_processor.media_type_json = "floober/blah"
    actual = harness.simple( "[---]\n[---] floober/blah\n{'Greetings': 'program!'}"
                           , filepath="foo.json.spt"
                            ).media_type
    assert actual == 'floober/blah'

def test_json_content_type_is_per_file_configurable(harness):
    expected = 'floober/blah'
    SPT="""
[---]
[---] floober/blah
{'Greetings': 'program!'}
"""
    actual = harness.simple(SPT, filepath="foo.json.spt").media_type
    assert actual == expected

def test_json_handles_unicode(harness):
    expected = b'''{
    "Greetings": "\\u00b5"
}'''
    actual = harness.simple('''
        from six import unichr
        [---]
        [---] application/json
        {'Greetings': unichr(181)}
    ''', filepath="foo.json.spt").body
    assert actual == expected

def test_json_doesnt_handle_non_ascii_bytestrings(harness):
    raises( (TypeError, UnicodeDecodeError)
          , harness.simple
          , "[---]\n[---] application/json\n{'Greetings': chr(181).encode('utf8')}"
          , filepath="foo.json.spt"
           )

def test_json_handles_time(harness):
    expected = '''{
    "seen": "19:30:00"
}'''
    actual = harness.simple('''
        [---]
        import datetime
        [---------------] application/json
        {'seen': datetime.time(19, 30)}
    ''', filepath="foo.json.spt").text
    assert actual == expected

def test_json_handles_date(harness):
    expected = '''{
    "created": "2011-05-09"
}'''
    actual = harness.simple('''

        [---]
        import datetime
        [---------------] application/json
        {'created': datetime.date(2011, 5, 9)}
    ''', filepath='foo.json.spt').text
    assert actual == expected

def test_json_handles_datetime(harness):
    expected = '''{
    "timestamp": "2011-05-09T00:00:00"
}'''
    actual = harness.simple("""
        [---]
        import datetime
        [---------------] application/json
        {'timestamp': datetime.datetime(2011, 5, 9, 0, 0)}
    """, filepath="foo.json.spt").text
    assert actual == expected

def test_json_handles_complex(harness):
    expected = '''{
    "complex": [
        1.0,
        2.0
    ]
}'''
    actual = harness.simple( "[---]\n[---] application/json\n{'complex': complex(1,2)}"
                           , filepath="foo.json.spt"
                            ).text
    # The json module puts trailing spaces after commas, but simplejson
    # does not. Normalize the actual input to work around that.
    actual = '\n'.join([line.rstrip() for line in actual.split('\n')])
    assert actual == expected

def test_json_raises_TypeError_on_unknown_types(harness):
    raises( TypeError
          , harness.simple
          , contents='[---]\nclass Foo: pass\n[---] application/json\nFoo()'
          , filepath='foo.json.spt'
           )

def test_aspen_json_load_loads():
    fp = io.StringIO()
    fp.write('{"cheese": "puffs"}')
    fp.seek(0)
    actual = json_.load(fp)
    assert actual == {'cheese': 'puffs'}

def test_aspen_json_dump_dumps():
    fp = io.BytesIO() if str is bytes else io.StringIO()
    json_.dump({"cheese": "puffs"}, fp)
    fp.seek(0)
    actual = fp.read()
    assert actual == '''{
    "cheese": "puffs"
}'''

def test_aspen_json_loads_loads():
    actual = json_.loads('{"cheese": "puffs"}')
    assert actual == {'cheese': 'puffs'}

def test_aspen_json_dumps_dumps():
    actual = json_.dumps({'cheese': 'puffs'})
    assert actual == '''{
    "cheese": "puffs"
}'''

# jsonp

JSONP_SIMPLATE = """[---]\n[---] application/json via jsonp_dump
{'Greetings': 'program!'}"""

JSONP_RESULT = '''/**/ foo({
    "Greetings": "program!"
});'''

def _jsonp_query(harness, querystring):
    return harness.simple(JSONP_SIMPLATE, filepath='index.json.spt', querystring=querystring).text

def test_jsonp_basically_works(harness):
    actual = _jsonp_query(harness, "jsonp=foo")
    assert actual == JSONP_RESULT, "wanted %r got %r " % (JSONP_RESULT, actual)

def test_jsonp_basically_works_with_callback(harness):
    actual = _jsonp_query(harness, "callback=foo")
    assert actual == JSONP_RESULT, "wanted %r got %r " % (JSONP_RESULT, actual)

def test_jsonp_defaults_to_json_with_no_callback(harness):
    expected = '''{
    "Greetings": "program!"
}'''
    actual = harness.simple(JSONP_SIMPLATE, filepath='index.spt').text
    assert actual == expected, "wanted %r got %r " % (expected, actual)

def test_jsonp_filters_disallowed_chars(harness):
    actual = _jsonp_query(harness, "callback=f+o+o")
    assert actual == JSONP_RESULT, "wanted %r got %r " % (JSONP_RESULT, actual)





#!/usr/bin/python

import os.path
import posixpath

import pytest

from aspen.http.request import Path
from aspen.request_processor.dispatcher import DispatchStatus
from aspen.testing import Harness

tablefile = os.path.join(os.path.dirname(__file__), 'dispatch_table_data.rst')

def find_cols(defline, header_char='='):
    """
    return a sorted list of (start, end) indexes into defline that
    are the beginning and ending indexes of column definitions
    based on a reStructuredText table header line.
    Note that this is a braindead simple version that only understands
    header_chars and spaces (no other whitespace)
    """
    i = 0;
    colstarts = []
    colends = []
    while i < len(defline):
        if len(colstarts) <= len(colends):
            nextstart = defline.find(header_char, i)
            if nextstart >= 0:
                colstarts.append(nextstart)
                i = nextstart
            else:
                break
        else:
            nextend = defline.find(' ',i)
            if nextend >= 0:
                colends.append(nextend)
                i = nextend
            else:
                colends.append(len(defline))
                break

    return list(zip(colstarts, colends))

def fields_from(dataline, cols):
    """
    Given a data line and a set of column definitions,
    strip the data and return it as a list
    """
    fields = []
    for start, fin in cols:
        fields.append(dataline[start:fin].strip())
    return fields

def get_table_entries():
    table = open(tablefile,'r').readlines()

    while table[0][:1] == '#':  # skip comment lines
        table = table[1:]

    tabledefline = table[0].strip()
    cols = find_cols(tabledefline)

    headers = fields_from(table[1], cols)
    inputfiles = headers[:headers.index('/')]
    requests = headers[headers.index('/'):]

    # We 'know' that table[0] == table[2], both header deflines, so skip down
    results = []
    for line in table[3:]:
        if line.strip() == tabledefline: break # found ending header, ignore the rest
        if line.strip().startswith('#'): continue # skip comment lines
        fields = fields_from(line, cols)
        files = [ x for i, x in enumerate(inputfiles) if fields[i] == 'X' ]
        expected = fields[len(inputfiles):]
        results += [ (files, r, expected[i])
                     for i, r in enumerate(requests) ]
    return results

def get_result(harness, request_uri):
    url_path = Path(request_uri)
    dispatch_result = harness.request_processor.dispatch(url_path)
    if dispatch_result.match:
        if dispatch_result.status == DispatchStatus.okay:
            result = 'ok'
        elif dispatch_result.status == DispatchStatus.unindexed:
            result = 'ui'
        else:
            result = str(dispatch_result.status)
        fspath = dispatch_result.match
        if os.sep != posixpath.sep:
            fspath = fspath.replace(os.sep, posixpath.sep)
        result += " " + (fspath[len(harness.fs.www.root)+1:] or '/')
        wilds = url_path
        if wilds:
            wildtext = ",".join("%s='%s'" % (k, wilds[k]) for k in sorted(wilds))
            result += " (%s)" % wildtext
        if dispatch_result.canonical:
            result += " @" + dispatch_result.canonical
    else:
        result = '-'
    return result

GENERIC_SPT = """
[-----]
[-----] text/plain
Greetings, Program!
"""

@pytest.mark.parametrize("files,request_uri,expected", get_table_entries())
def test_all_table_entries(harness, files, request_uri, expected):
    # set up the specified files
    realfiles = tuple([ f if f.endswith('/') else (f, GENERIC_SPT) for f in files ])
    harness.fs.www.mk(*realfiles)
    result = get_result(harness, request_uri)
    assert result == expected, "Requesting %r, got %r instead of %r" % (request_uri, result, expected)


if __name__ == '__main__':
    # output the table with answers the current dispatcher gives
    # currently this has to be run manually with:
    #    ./env/bin/python tests/dispatch_table_test | grep -v ^pid
    table = open(tablefile,'r').readlines()

    tabledefline = table[0].strip()
    cols = find_cols(tabledefline)

    headers = fields_from(table[1], cols)
    answercol = headers.index('/')
    inputfiles = headers[:answercol]
    requests = headers[answercol:]

    for line in table[:3]:
        print(line)
    # We 'know' that table[0] == table[2], both header deflines, so skip down
    for line in table[3:]:
        if line.strip() == tabledefline: break # found ending header, ignore the rest
        if line.strip().startswith('#'): continue # skip comment lines
        fields = fields_from(line, cols)
        files = [ x for i, x in enumerate(inputfiles) if fields[i] == 'X' ]
        expected = fields[len(inputfiles):]
        resultcolstart = cols[answercol][0]
        resultline = line[:resultcolstart] # input files
        harness = Harness()
        realfiles = tuple([ f if f.endswith('/') else (f, GENERIC_SPT) for f in files ])
        harness.fs.www.mk(*realfiles)
        for i, request_uri in enumerate(requests):
            result = get_result(harness, request_uri)
            col = answercol + i
            resultline += result + (' ' * (cols[col][1] - cols[col][0] - len(result)))
            if col < len(cols) - 1:
                resultline += ' ' * (cols[col+1][0] - cols[col][1])
        print(resultline)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import sys

import pytest
from aspen.testing import Harness, teardown


@pytest.fixture
def sys_path_scrubber():
    before = set(sys.path)
    yield
    after = set(sys.path)
    for name in after - before:
        sys.path.remove(name)


@pytest.fixture
def harness(sys_path_scrubber):
    harness = Harness()
    yield harness
    harness.teardown()


def pytest_runtest_teardown():
    teardown()



# encoding: utf8
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pytest import raises

from aspen.http.mapping import Mapping


def test_mapping_subscript_assignment_clobbers():
    m = Mapping()
    m['foo'] = 'bar'
    m['foo'] = 'baz'
    m['foo'] = 'buz'
    expected = ['buz']
    actual = dict.__getitem__(m, 'foo')
    assert actual == expected

def test_mapping_subscript_access_returns_last():
    m = Mapping()
    m['foo'] = 'bar'
    m['foo'] = 'baz'
    m['foo'] = 'buz'
    expected = 'buz'
    actual = m['foo']
    assert actual == expected

def test_mapping_get_returns_last():
    m = Mapping()
    m['foo'] = 'bar'
    m['foo'] = 'baz'
    m['foo'] = 'buz'
    expected = 'buz'
    actual = m.get('foo')
    assert actual == expected

def test_mapping_get_returns_default():
    m = Mapping()
    expected = 'cheese'
    actual = m.get('foo', 'cheese')
    assert actual == expected

def test_mapping_get_default_default_is_None():
    m = Mapping()
    expected = None
    actual = m.get('foo')
    assert actual is expected

def test_mapping_all_returns_list_of_all_values():
    m = Mapping()
    m['foo'] = 'bar'
    m.add('foo', 'baz')
    m.add('foo', 'buz')
    expected = ['bar', 'baz', 'buz']
    actual = m.all('foo')
    assert actual == expected

def test_mapping_all_returns_empty_list_when_key_is_missing():
    m = Mapping()
    expected = []
    actual = m.all('foo')
    assert actual == expected

def test_mapping_ones_returns_list_of_last_values():
    m = Mapping()
    m['foo'] = 1
    m['foo'] = 2
    m['bar'] = 3
    m['bar'] = 4
    m['bar'] = 5
    m['baz'] = 6
    m['baz'] = 7
    m['baz'] = 8
    m['baz'] = 9
    expected = [2, 5, 9]
    actual = m.ones('foo', 'bar', 'baz')
    assert actual == expected

def test_mapping_deleting_a_key_removes_it_entirely():
    m = Mapping()
    m['foo'] = 1
    m['foo'] = 2
    m['foo'] = 3
    del m['foo']
    assert 'foo' not in m

def test_accessing_missing_key_calls_keyerror():
    m = Mapping()
    class Foobar(Exception): pass
    def raise_foobar(self):
        raise Foobar
    m.keyerror = raise_foobar
    raises(Foobar, lambda k: m[k], 'foo')
    raises(Foobar, m.ones, 'foo')

def test_mapping_pop_returns_the_last_item():
    m = Mapping()
    m['foo'] = 1
    m.add('foo', 1)
    m.add('foo', 3)
    expected = 3
    actual = m.pop('foo')
    assert actual == expected

def test_mapping_pop_leaves_the_rest():
    m = Mapping()
    m['foo'] = 1
    m.add('foo', 1)
    m.add('foo', 3)
    m.pop('foo')
    expected = [1, 1]
    actual = m.all('foo')
    assert actual == expected

def test_mapping_pop_removes_the_item_if_that_was_the_last_value():
    m = Mapping()
    m['foo'] = 1
    m.pop('foo')
    expected = []
    actual = list(m.keys())
    assert actual == expected

def test_mapping_pop_raises_KeyError_by_default():
    m = Mapping()
    with raises(KeyError):
        m.pop('foo')

def test_mapping_popall_returns_a_list():
    m = Mapping()
    m['foo'] = 1
    m.add('foo', 1)
    m.add('foo', 3)
    expected = [1, 1, 3]
    actual = m.popall('foo')
    assert actual == expected

def test_mapping_popall_removes_the_item():
    m = Mapping()
    m['foo'] = 1
    m['foo'] = 1
    m['foo'] = 3
    m.popall('foo')
    assert 'foo' not in m



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


NO_DEFAULT = object()


class Mapping(dict):
    """Base class for HTTP mappings.

    Mappings in HTTP differ from Python dictionaries in that they may have one
    or more values. This dictionary subclass maintains a list of values for
    each key. However, access semantics are asymmetric: subscript assignment
    clobbers to list, while subscript access returns the last item. Think
    about it.

    .. warning:: This isn't thread-safe.

    """

    def __getitem__(self, name):
        """Given a name, return the last value or call self.keyerror.
        """
        try:
            return dict.__getitem__(self, name)[-1]
        except KeyError:
            self.keyerror(name)

    def __setitem__(self, name, value):
        """Given a name and value, clobber any existing values.
        """
        dict.__setitem__(self, name, [value])

    def keyerror(self, key):
        """Called when a key is missing. Default implementation simply reraises.
        """
        raise

    def pop(self, name, default=NO_DEFAULT):
        """Given a name, return a value.

        This removes the last value from the list for name and returns it. If
        there was only one value in the list then the key is removed from the
        mapping. If name is not present and default is given, that is returned
        instead. Otherwise, self.keyerror is called.

        """
        try:
            values = dict.__getitem__(self, name)
        except KeyError:
            if default is not NO_DEFAULT:
                return default
            self.keyerror(name)
        value = values.pop()
        if not values:
            del self[name]
        return value

    popall = dict.pop

    def all(self, name):
        """Given a name, return a list of values, possibly empty.
        """
        return dict.get(self, name, [])

    def get(self, name, default=None):
        """Override to only return the last value.
        """
        return dict.get(self, name, [default])[-1]

    def add(self, name, value):
        """Given a name and value, clobber any existing values with the new one.
        """
        if name in self:
            self.all(name).append(value)
        else:
            dict.__setitem__(self, name, [value])

    def ones(self, *names):
        """Given one or more names of keys, return a list of their values.
        """
        lowered = []
        for name in names:
            n = name.lower()
            if n not in lowered:
                lowered.append(n)
        return [self[name] for name in lowered]



"""
This module defines all of the custom exceptions used across the Aspen library.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


class ConfigurationError(Exception):
    """This is an error in any part of our configuration.
    """

    def __init__(self, msg):
        Exception.__init__(self)
        self.msg = msg

    def __str__(self):
        return self.msg


class NegotiationFailure(Exception):
    """The requested media type isn't available (HTTP status code 406).
    """

    def __init__(self, accept, available_types):
        self.accept = accept
        self.available_types = available_types
        message = "Couldn't satisfy %s. The following media types are available: %s."
        self.message = message % (self.accept, ', '.join(self.available_types))

    def __str__(self):
        return self.message


class TypecastError(Exception):
    """Parsing a segment of the request path failed (HTTP status code 404).
    """

    def __init__(self, extension):
        self.msg = "Failure to typecast extension '{0}'".format(extension)
        Exception.__init__(self)


class NotFound(Exception):
    """The requested resource isn't available (HTTP status code 404).
    """

    def __init__(self, message=''):
        self.message = message or "not found"


class AttemptedBreakout(Exception):
    """Raised when a request is dispatched to a symlinked file which is outside
    :attr:`~aspen.request_processor.DefaultConfiguration.www_root`.
    """

    def __init__(self, sym_path, real_path):
        self.sym_path = sym_path
        self.real_path = real_path

    def __str__(self):
        if self.real_path == self.sym_path:
            return "%r isn't inside a known resource directory" % self.sym_path
        return "%r is a symlink to %r" % (self.sym_path, self.real_path)


class PossibleBreakout(AttemptedBreakout, Warning):
    """A warning emitted when a symlink that points outside
    :attr:`~aspen.request_processor.DefaultConfiguration.www_root` is detected.
    """


class SlugCollision(Exception):
    """Raised if two files claim the same URL path.

    Example: ``foo.html`` and ``foo.html.spt`` both claim ``/foo.html``.
    """

    def __init__(self, slug, node1, node2):
        self.slug = slug
        self.node1 = node1
        self.node2 = node2

    def __str__(self):
        return (
            "The URL segment %r is claimed by two different nodes:\n"
            "%s(fspath=%r, type=%r, ...) and\n"
            "%s(fspath=%r, type=%r, ...)"
        ) % (
            self.slug,
            self.node1.__class__.__name__, self.node1.fspath, self.node1.type,
            self.node2.__class__.__name__, self.node2.fspath, self.node2.type,
        )

class WildcardCollision(Exception):
    """Raised if a filesystem path contains multiple wildcards with the same name.

    Examples: ``www/%foo/%foo/index.spt``, ``www/%foo/bar/%foo.spt``.
    """

    def __init__(self, varname, fspath):
        self.varname = varname
        self.fspath = fspath

    def __str__(self):
        return "%r appears twice in the filesystem path %r" % ('%' + self.varname, self.fspath)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from io import BytesIO
import re

from six import PY2

from ..output import Output
from .pagination import split_and_escape, parse_specline, Page
from aspen.http.resource import Dynamic, open_resource


DEFAULT_ENCODING = 'ascii' if PY2 else 'utf8'


renderer_re = re.compile(r'[a-z0-9.-_]+$')
media_type_re = re.compile(r'[A-Za-z0-9.+*-]+/[A-Za-z0-9.+*-]+$')


def _decode(raw):
    """Decode the raw bytes of a simplate.

    As per PEP 263, decode raw data according to the encoding specified in the
    first couple lines of the data, or using the default encoding (ASCII in
    Python 2, UTF-8 in Python 3 per PEP 3120).

    Raises: :class:`UnicodeDecodeError` if decoding fails.
    """
    assert type(raw) is bytes  # sanity check

    decl_re = re.compile(br'^[ \t\f]*#.*coding[:=][ \t]*([-\w.]+)')

    def get_declaration(line):
        match = decl_re.match(line)
        if match:
            return match.group(1)
        return None

    encoding = None
    fulltext = b''
    sio = BytesIO(raw)
    for line in (sio.readline(), sio.readline()):
        potential = get_declaration(line)
        if potential is not None:
            if encoding is None:

                # If both lines match, use the first. This matches Python's
                # observed behavior.

                encoding = potential
                munged = b'# encoding set to ' + encoding + b'\n'

            else:

                # But always munge any encoding line. We can't simply remove
                # the line, because we want to preserve the line numbering.
                # However, later on when we ask Python to exec a unicode
                # object, we'll get a SyntaxError if we have a well-formed
                # `coding: # ` line in it.

                munged = b'# encoding NOT set to ' + potential + b'\n'

            line = line.split(b'#')[0] + munged

        fulltext += line
    fulltext += sio.read()
    sio.close()
    encoding = encoding.decode('ascii') if encoding else DEFAULT_ENCODING
    return fulltext.decode(encoding)


class SimplateDefaults(object):
    def __init__(self, renderers_by_media_type, renderer_factories, initial_context):
        """
        Things that are usually the same across all simplates:

        renderers_by_media_type - dict[media_type_name] -> renderer_name
        renderer_factories - dict[renderer_name] -> renderer_factory
        initial_context - initial context passed into the 'run-once' page
        """
        self.renderers_by_media_type = renderers_by_media_type # type: Dict[str, str]
        self.renderer_factories = renderer_factories           # type: Dict[str, Callable]
        self.initial_context = initial_context                 # type: Dict[str, object]


class Simplate(Dynamic):
    """A simplate is a dynamic resource with multiple syntaxes in one file.

    Args:
        fspath (str): the absolute filesystem path of this simplate

    """

    __slots__ = (
        'fspath', 'default_media_type', 'renderers', 'page_one', 'page_two',
    )

    defaults = None # type: SimplateDefaults

    def __init__(self, request_processor, fspath):
        self.request_processor = request_processor
        self.fspath = fspath
        self.default_media_type = request_processor.guess_media_type(fspath.rsplit('.', 1)[0])

        self.renderers = {}         # mapping of media type to Renderer objects
        self.available_types = []   # ordered sequence of media types
        with open_resource(request_processor, fspath) as fh:
            raw = fh.read()
        pages = self.parse_into_pages(_decode(raw))
        self.compile_pages(pages)
        self.page_one, self.page_two = pages[0], pages[1]
        for renderer, media_type in pages[2:]:
            if media_type in self.renderers:
                raise SyntaxError("Two content pages defined for %s." % media_type)
            self.available_types.append(media_type)
            self.renderers[media_type] = renderer


    def render_for_type(self, media_type, context):
        """Render the simplate.

        Args:
            media_type (str): the media type of the page to render
            context (dict): execution context values you wish to supply

        .. warning: The ``context`` dict is updated with values from the first
                    and second pages of the simplate.

        Returns: an :class:`Output` object.
        """

        # create Output object and put it in the context
        output = context['output'] = Output(media_type=media_type)
        # update the context with values from the first page
        context.update(self.page_one)
        # use this as the context to execute the second page in
        exec(self.page_two, context)
        # skip rendering if the second page has already filled output.body
        if output.body is not None:
            return output

        if '__all__' in context:
            # templates will only see variables named in __all__
            context = dict([ (k, context[k]) for k in context['__all__'] ])

        # load the renderer
        render = self.renderers[media_type]
        # render
        output.body = render(context)

        return output


    def parse_into_pages(self, decoded):
        """Given a bytestring that is the entire simplate, return a list of pages.

        If there's one page, it's a template.

        If there's more than one page, the first page is always python and the last is always a template.

        If there's more than two pages, the second page is python *unless it has a specline*, which makes it a template.

        """

        pages = list(split_and_escape(decoded))
        npages = len(pages)
        blank = [ Page('') ]

        if npages == 1:
            pages = blank + blank + pages
        elif npages == 2:
            pages = blank + pages
        elif pages[1].header: # it's got a header, so it's a template
            pages = blank + pages

        return pages


    def compile_pages(self, pages):
        """Given a list of pages, replace the pages with objects.

        Page 0 is the 'run once' page - it is executed and the resulting
        context stored in :obj:`self.pages[0]`.

        Page 1 is the 'run every' page - it is compiled for easier execution
        later, and stored in :obj:`self.pages[1]`.

        Subsequent pages are templates, so each one's content type and
        respective renderer are stored as a tuple in :obj:`self.pages[n]`.
        """

        # Exec the first page and compile the second.
        # ===========================================

        one, two = pages[:2]

        context = dict()
        context['__file__'] = self.fspath
        context.update(self.defaults.initial_context)

        one = compile(one.padded_content, self.fspath, 'exec')
        exec(one, context)    # mutate context
        one = context          # store it

        two = compile(two.padded_content, self.fspath, 'exec')

        pages[:2] = (one, two)
        pages[2:] = [self.compile_page(page) for page in pages[2:]]


    def compile_page(self, page):
        """Given a :class:`Page`, return a :obj:`(renderer, media_type)` pair.
        """
        make_renderer, media_type = self._parse_specline(page.header)
        renderer = make_renderer(self.fspath, page.content, media_type, page.offset)
        return (renderer, media_type)

    def _parse_specline(self, specline):
        """Given a bytestring, return a two-tuple.

        The incoming string is expected to be of the form:

            media_type via renderer

        Both are optional.

        The media_type will default to the default_media_type supplied to
        this simplate at instantiation time.  (Possibly determined by a
        file extension or other metadata)

        The renderer will be computed based on media type if absent.

        The return two-tuple contains a render function and a media
        type (as unicode). SyntaxError is raised if there aren't one or two
        parts or if either of the parts is malformed. If only one part is
        passed in it's interpreted as a media type.

        """
        # Parse into parts
        media_type, renderer = parse_specline(specline)

        if media_type == '':
            # no media type specified, use the default
            media_type = self.default_media_type
        if renderer == '':
            # no renderer specified, use the default
            renderer = self.defaults.renderers_by_media_type[media_type]

        # Validate media type.
        if media_type_re.match(media_type) is None:
            msg = ("Malformed media type '%s' in specline '%s'. It must match "
                   "%s.")
            msg %= (media_type, specline, media_type_re.pattern)
            raise SyntaxError(msg)

        # Hydrate and validate renderer.
        make_renderer = self._get_renderer_factory(media_type, renderer)

        # Return.
        return (make_renderer, media_type)


    def _get_renderer_factory(self, media_type, renderer):
        """Given two bytestrings, return a renderer factory or None.
        """
        factories = self.defaults.renderer_factories
        if renderer_re.match(renderer) is None:
            possible =', '.join(sorted(factories.keys()))
            msg = ("Malformed renderer %s. It must match %s. Possible "
                   "renderers (might need third-party libs): %s.")
            raise SyntaxError(msg % (renderer, renderer_re.pattern, possible))

        make_renderer = factories.get(renderer, None)
        if isinstance(make_renderer, ImportError):
            raise make_renderer
        elif make_renderer is None:
            possible = []
            legend = ''
            for k, v in sorted(factories.items()):
                if isinstance(v, ImportError):
                    k = '*' + k
                    legend = " (starred are missing third-party libraries)"
                possible.append(k)
            possible = ', '.join(possible)
            raise ValueError("Unknown renderer for %s: %s. Possible "
                             "renderers%s: %s."
                             % (media_type, renderer, legend, possible))
        return make_renderer



"""
This module implements finding the file that matches a request path.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from functools import reduce
from inspect import isclass
from operator import attrgetter
import os
import posixpath
import warnings

try:
    from os import scandir
except ImportError:
    from scandir import scandir

from ..exceptions import PossibleBreakout, SlugCollision, WildcardCollision

from ..utils import auto_repr, Constant


def debug_noop(msg, *args):
    pass


def debug_stdout(msg, *args):
    r = msg(*args) if callable(msg) else msg % args
    try:
        print("DEBUG: " + r)
    except Exception:
        print("DEBUG: " + repr(r))

ASPEN_DEBUG = 'ASPEN_DEBUG' in os.environ
debug = debug_stdout if ASPEN_DEBUG else debug_noop


def get_mtime_ns(fspath):
    # For compatibility with Python < 3.3
    st = os.stat(fspath)
    return getattr(st, 'st_mtime_ns', int(st.st_mtime * 1000000000))


def splitext(name):
    return name.rsplit('.', 1) if '.' in name else [name, None]


def strip_matching_ext(a, b):
    """Given two names, strip a trailing extension iff they both have them.
    """
    aparts = splitext(a)
    bparts = splitext(b)

    def debug_ext():
        return "exts: %r( %r ) and %r( %r )" % (a, aparts[1], b, bparts[1])

    if aparts[1] == bparts[1]:
        debug(lambda: debug_ext() + " matches")
        return aparts[0], bparts[0]
    debug(lambda: debug_ext() + " don't match")
    return a, b


class DispatchStatus(object):
    """The attributes of this class are constants that represent dispatch statuses."""

    okay = Constant('okay')
    "Found a matching file."

    missing = Constant('missing')
    "No match found."

    unindexed = Constant('unindexed')
    "Found a matching node, but it's a directory without an index."


@auto_repr
class DispatchResult(object):
    """The result of a dispatch operation."""

    __slots__ = ('status', 'match', 'wildcards', 'extension', 'canonical')

    def __init__(self, status, match, wildcards, extension, canonical):
        self.status = status
        "A :class:`DispatchStatus` constant encoding the overall result."

        self.match = match
        "The matching filesystem path (if status != 'missing')."

        self.wildcards = wildcards or None
        "A dict whose keys are wildcard names, and values are as supplied by the path."

        self.extension = extension
        "A file extension, e.g. ``json`` when ``foo.spt`` is matched to ``foo.json``."

        self.canonical = canonical
        "The canonical path of the resource, e.g. ``/`` for ``/index.html``."

    def __eq__(self, other):
        return (
            isinstance(other, DispatchResult) and
            self.status == other.status and
            self.match == other.match and
            self.wildcards == other.wildcards and
            self.extension == other.extension and
            self.canonical == other.canonical
        )

    def __ne__(self, other):
        return not (self == other)

    def _as_tuple(self):
        # We can't give this class a proper `__hash__` method because
        # `DispatchResult` objects aren't immutable, so we use this method
        # instead to get a hashable tuple.
        return (
            self.status,
            self.match,
            frozenset(self.wildcards.items()) if self.wildcards else self.wildcards,
            self.extension,
            self.canonical
        )


MISSING = DispatchResult(DispatchStatus.missing, None, None, None, None)


@auto_repr
class FileNode(object):
    """Represents a file in a dispatch tree."""

    __slots__ = ('fspath', 'type', 'wildcard', 'extension')

    def __init__(self, fspath, type, wildcard, extension):
        self.fspath = fspath
        "The absolute filesystem path of this node."

        self.type = type
        "The node's type: 'dynamic' or 'static'."

        self.wildcard = wildcard
        "The name of the path variable if the node is a wildcard."

        self.extension = extension
        "The sub-extension of a dynamic file, e.g. ``json`` for ``foo.json.spt``."


@auto_repr
class DirectoryNode(object):
    """Represents a directory in a dispatch tree."""

    __slots__ = ('fspath', 'wildcard', 'children')

    type = 'directory'

    def __init__(self, fspath, wildcard, children):
        self.fspath = fspath
        "The absolute filesystem path of this node."

        self.wildcard = wildcard
        "The name of the path variable if the node is a wildcard."

        self.children = children
        "The node's children as a dict (keys are names and values are nodes)."


@auto_repr
class LiveDirectoryNode(object):
    """Dynamically represents a directory in a dispatch tree."""

    __slots__ = ('fspath', 'wildcard', '_children', 'mtime', 'dispatcher')

    type = 'directory'

    def __init__(self, fspath, wildcard, children, mtime, dispatcher):
        self.fspath = fspath
        "The absolute filesystem path of this node."

        self.wildcard = wildcard
        "The name of the path variable if the node is a wildcard."

        self._children = children
        "The node's children as a dict (keys are names and values are nodes)."

        self.mtime = mtime
        "The last modification time of the directory, in nanoseconds."

        self.dispatcher = dispatcher
        "Points to the :class:`Dispatcher` object that created this node."

    @property
    def children(self):
        mtime = get_mtime_ns(self.fspath)
        if mtime != self.mtime:
            dirnames = self.fspath[len(self.dispatcher.www_root)+1:].split(os.path.sep)
            varnames = {
                dirname[1:]: os.path.sep.join([self.dispatcher.www_root] + dirnames[:i+1])
                for i, dirname in enumerate(dirnames)
                if dirname.startswith('%')
            }
            self._children, self.mtime = self.dispatcher._build_subtree(self.fspath, varnames)
        return self._children


# Collision handlers
# ==================

def legacy_collision_handler(slug, node1, node2):
    """Ignores all collisions, like :class:`SystemDispatcher` does.
    """
    if node1.type == 'directory' and node2.type != 'directory':
        if not node1.children:
            # Ignore empty directory
            return 'replace_first_node'
        if '' not in node1.children:
            # Allow `/bar.spt` to act as the index of `/bar/`
            return 'set_second_node_as_index_of_first_node'
    return 'ignore_second_node'


def strict_collision_handler(*args):
    """A sane collision handler, it doesn't allow any.
    """
    return 'raise'


def hybrid_collision_handler(slug, node1, node2):
    """This collision handler allows a static file to shadow a dynamic resource.

    Example: ``/file.js`` will be preferred over ``/file.js.spt``.
    """
    if node1.type == 'directory' and node2.type != 'directory':
        if not node1.children:
            # Ignore empty directory
            return 'replace_first_node'
        if '' not in node1.children:
            # Allow `/bar.spt` to act as the index of `/bar/`
            return 'set_second_node_as_index_of_first_node'
    elif node1.type == 'static' and node2.type == 'dynamic':
        # Allow `/foo.css` to shadow `/foo.css.spt`
        return 'ignore_second_node'
    return 'raise'


# File skippers
# =============

def skip_hidden_files(name, dirpath):
    """Skip all names starting with a dot, except ``.well-known``.
    """
    return name[0] == '.' and name != '.well-known'


def skip_nothing(name, dirpath):
    """Always returns :obj:`False`.
    """
    return False


# Dispatcher classes
# ==================

class Dispatcher(object):
    """The abstract base class of dispatchers.

    Args
    ----
    www_root
        the path to a filesystem directory
    is_dynamic
        a function that takes a file name and returns a boolean
    indices
        a list of filenames that should be treated as directory indexes
    typecasters
        a dict of typecasters, keys are strings and values are functions
    file_skipper
        a function that takes a file name and a directory path and returns a boolean
    collision_handler
        a function that takes 3 arguments (`slug, node1, node2`) and returns a string
    """

    def __init__(
        self, www_root, is_dynamic, indices, typecasters,
        file_skipper=skip_hidden_files, collision_handler=hybrid_collision_handler
    ):
        self.www_root = os.path.realpath(www_root)
        self.is_dynamic = is_dynamic
        self.indices = indices
        self.typecasters = typecasters
        self.file_skipper = file_skipper
        self.collision_handler = collision_handler

    def build_dispatch_tree(self):
        """Called to build the dispatch tree.

        Subclasses **must** implement this method.
        """
        raise NotImplementedError('abstract method')

    def dispatch(self, path, path_segments):
        """Dispatch a request.

        Args:
            path (str): the request path, e.g. ``'/'``
            path_segments (list): the path split into segments, e.g. ``['']``

        Subclasses **must** implement this method.
        """
        raise NotImplementedError('abstract method')

    def find_index(self, dirpath):
        """Looks for an index file in a directory.

        Returns:
            the full path of the first index file, or :obj:`None` if no index was found
        """
        for filename in self.indices:
            index = os.path.join(dirpath, filename)
            if os.path.isfile(index):
                return index
        return None

    def split_wildcard(self, wildcard, is_dir):
        """Splits a wildcard into its components.

        Args:
            wildcard (str): the string to split, e.g. :obj:`'year.int'`
            is_dir (bool): :obj:`True` if the wildcard is from a directory name

        Returns:
            a 3-tuple ``(varname, vartype, extension)``
        """
        if '.' not in wildcard:
            return wildcard, None, None
        if is_dir:
            extension = None
            varname, vartype = wildcard.rsplit('.', 1)
            if vartype not in self.typecasters:
                varname, vartype = wildcard, None
        else:
            try:
                varname, vartype, extension = wildcard.split('.', 2)
            except ValueError:
                varname, ambiguous = wildcard.split('.')
                if ambiguous in self.typecasters:
                    vartype, extension = ambiguous, None
                else:
                    vartype, extension = None, ambiguous
                del ambiguous
        return varname, vartype, extension


class SystemDispatcher(Dispatcher):
    """Aspen's original dispatcher, it's very inefficient.
    """

    def build_dispatch_tree(self):
        """"""
        pass

    def dispatch(self, path, path_segments):
        """"""
        listnodes = os.listdir
        is_leaf = os.path.isfile
        traverse = os.path.join
        result = _dispatch_abstract(
            self, listnodes, self.is_dynamic, is_leaf, traverse, self.find_index,
            self.www_root, path_segments
        )
        debug(lambda: "dispatch_abstract returned: " + repr(result))
        return result


def _dispatch_abstract(dispatcher, listnodes, is_dynamic, is_leaf, traverse, find_index, startnode, nodepath):
    """Given a list of nodenames (in 'nodepath'), return a DispatchResult.

    We try to traverse the directed graph rooted at 'startnode' using the
    functions:

       listnodes(joinedpath) - lists the nodes in the specified joined path

       is_dynamic(node) - returns true iff the specified node is dynamic

       is_leaf(node) - returns true iff the specified node is a leaf node

       traverse(joinedpath, newnode) - returns a new joined path by traversing
        into newnode from the current joinedpath

       find_index(joinedpath) - returns the index file in the specified path if
        it exists, or None if not

    Wildcards nodenames start with %. Non-leaf wildcards are used as keys in
    wildvals and their actual path names are used as their values. In general,
    the rule for matching is 'most specific wins': $foo looks for isfile($foo)
    then isfile($foo-minus-extension) then isfile(virtual-with-extension) then
    isfile(virtual-no-extension) then isdir(virtual)

    """
    file_skipper = dispatcher.file_skipper

    nodepath = nodepath[:]  # copy it so we can mutate it if necessary
    wildvals, wildleafs = {}, {}
    curnode = startnode
    subnodes = None
    extension, canonical = None, None
    is_dynamic_node = lambda n: is_dynamic(traverse(curnode, n))
    is_leaf_node = lambda n: is_leaf(traverse(curnode, n))

    def get_wildleaf_fallback():
        lastnode_ext = splitext(nodepath[-1])[1]
        wildleaf_fallback = lastnode_ext in wildleafs or None in wildleafs
        if wildleaf_fallback:
            ext = lastnode_ext if lastnode_ext in wildleafs else None
            curnode, wildvals = wildleafs[ext]
            debug(lambda: "Wildcard leaf match %r and ext %r" % (curnode, ext))
            return DispatchResult(DispatchStatus.okay, curnode, wildvals, None, None)
        return None

    for depth, node in enumerate(nodepath):

        # check all the possibilities:
        # node.html, node.html.spt, node.spt, node.html/, %node.html/ %*.html.spt, %*.spt

        parent_subnodes, subnodes = subnodes, set()
        wild_leaf_ns, wild_nonleaf_ns = [], []
        for entry in scandir(curnode):
            if file_skipper(entry.name, curnode):
                # don't serve hidden files
                continue
            if entry.is_symlink():
                real_path = os.path.realpath(entry.path)
                if not real_path.startswith(startnode):
                    # don't serve files outside `www_root`
                    warnings.warn(PossibleBreakout(entry.path, real_path))
                    continue
            subnodes.add(entry.name)
            if entry.name.startswith("%"):
                if entry.is_dir():
                    wild_nonleaf_ns.append(entry.name)
                elif is_dynamic(entry.path):
                    wild_leaf_ns.append(entry.name)
        wild_leaf_ns.sort()
        wild_nonleaf_ns.sort()

        node_noext, node_ext = splitext(node)

        # store all the fallback possibilities
        remaining = reduce(posixpath.join, nodepath[depth:])
        for n in wild_leaf_ns:
            wildwildvals = wildvals.copy()
            varname, vartype, leaf_ext = dispatcher.split_wildcard(splitext(n[1:])[0], False)
            wildcard = '.'.join((varname, vartype)) if vartype else varname
            if leaf_ext and remaining.endswith(leaf_ext):
                wildwildvals[wildcard] = remaining[:-len(leaf_ext)-1]
            else:
                wildwildvals[wildcard] = remaining
            wildleafs[leaf_ext] = (traverse(curnode, n), wildwildvals)

        debug(lambda: "wildleafs is %r" % wildleafs)

        found_n = None
        last_node = (depth + 1) == len(nodepath)
        if last_node:
            debug(lambda: "on last node %r" % node)
            if node == '':  # dir request
                debug(lambda: "...last node is empty")
                path_so_far = traverse(curnode, node)
                index = find_index(path_so_far)
                if index:
                    debug(lambda: "found index: %r" % index)
                    return DispatchResult(DispatchStatus.okay, index, wildvals, None, canonical)
                if depth > 0 and nodepath[-2] + ".spt" in parent_subnodes:
                    if '.' not in nodepath[-2]:
                        debug(lambda: "slashless match")
                        curnode = reduce(traverse, nodepath[:-1], startnode) + ".spt"
                        if not subnodes:
                            # The directory is empty, so the canonical path is
                            # without the final slash
                            canonical = '/' + '/'.join(nodepath[:-1])
                        break
                if wild_leaf_ns:
                    found_n = wild_leaf_ns[0]
                    debug(lambda: "found wild leaf: %r" % found_n)
                    curnode = traverse(curnode, found_n)
                    varname, vartype, _ = dispatcher.split_wildcard(splitext(found_n[1:])[0], False)
                    wildcard = '.'.join((varname, vartype)) if vartype else varname
                    wildvals[wildcard] = node
                    return DispatchResult(DispatchStatus.okay, curnode, wildvals, None, canonical)
                debug(lambda: "no match")
                return DispatchResult(
                    DispatchStatus.unindexed, curnode + os.path.sep, wildvals, None, canonical
                )
            elif node in subnodes and is_leaf_node(node):
                debug(lambda: "...found exact file, must be static")
                if is_dynamic_node(node):
                    return MISSING
                else:
                    found_n = node
                    if find_index(curnode) == traverse(curnode, node):
                        # The canonical path of `/index.html` is `/`
                        canonical = '/' + '/'.join(nodepath)[:-len(node)]
            elif node + ".spt" in subnodes and is_leaf_node(node + ".spt"):
                debug(lambda: "...found exact spt")
                found_n = node + ".spt"
            elif node_noext + ".spt" in subnodes and is_leaf_node(node_noext + ".spt") \
                    and node_ext:
                # node has an extension
                debug(lambda: "...found indirect spt, extension is `%s`" % node_ext)
                # indirect match - foo.spt is answering to foo.html
                extension = node_ext
                found_n = node_noext + ".spt"

            if found_n is not None:
                debug(lambda: "found_n: %r" % found_n)
                curnode = traverse(curnode, found_n)
            elif wild_nonleaf_ns:
                debug(lambda: "wild_nonleaf_ns")
                result = get_wildleaf_fallback()
                if result:
                    return result
                found_n = wild_nonleaf_ns[0]
                wildvals[found_n[1:]] = node
                curnode = traverse(curnode, found_n)
                nodepath.append('')
                canonical = '/' + '/'.join(nodepath)
            elif node in subnodes:
                debug(lambda: "exact dirmatch")
                curnode = traverse(curnode, node)
                nodepath.append('')
                canonical = '/' + '/'.join(nodepath)
            else:
                debug(lambda: "fallthrough")
                result = get_wildleaf_fallback()
                if not result:
                    return MISSING
                return result

        if not last_node:  # not at last path seg in request
            debug(lambda: "on node %r" % node)
            if node in subnodes and not is_leaf_node(node):
                found_n = node
                debug(lambda: "Exact match " + repr(node))
                curnode = traverse(curnode, found_n)
            elif '.' not in node and node + ".spt" in subnodes and nodepath[depth+1:] == ['']:
                found_n = node + ".spt"
                debug(lambda: "Slashless match " + repr(found_n))
                curnode = traverse(curnode, found_n)
                canonical = '/' + '/'.join(nodepath[:-1])
                break
            elif wild_nonleaf_ns:
                # need to match a wildnode, and we're not the last node, so we should match
                # non-leaf first, then leaf
                found_n = wild_nonleaf_ns[0]
                wildvals[found_n[1:]] = node
                debug(lambda: "Wildcard match %r = %r " % (found_n, node))
                curnode = traverse(curnode, found_n)
            else:
                debug(lambda: "No exact match for " + repr(node))
                result = get_wildleaf_fallback()
                if not result:
                    return MISSING
                return result

    return DispatchResult(DispatchStatus.okay, curnode, wildvals, extension, canonical)


class UserlandDispatcher(Dispatcher):
    """A dispatcher optimized for production use.

    This dispatcher builds a complete and static tree when it is first created.
    It then uses this dispatch tree to route requests without making any system
    call, thus avoiding FFI and context switching costs.

    This is the default dispatcher (when the ``changes_reload`` configuration
    option is ``False``).
    """

    DIR_WILDCARD = Constant('DIR_WILDCARD')
    LEAF_WILDCARDS = Constant('LEAF_WILDCARDS')

    def make_dir_node(self, fspath, wildcard, children, mtime):
        return DirectoryNode(fspath, wildcard, children)

    def build_dispatch_tree(self):
        """"""
        children, mtime = self._build_subtree(self.www_root, {})
        self.tree = self.make_dir_node(self.www_root, None, children, mtime)

    def _build_subtree(self, dirpath, varnames):
        """This method recursively builds a dispacth subtree.
        """
        children = {}
        mtime = get_mtime_ns(dirpath)
        index = self.find_index(dirpath)
        for entry in sorted(scandir(dirpath), key=attrgetter('name')):
            name = entry.name
            if self.file_skipper(name, dirpath):
                continue
            fspath = entry.path
            if entry.is_symlink():
                fspath = os.path.realpath(fspath)
                if not fspath.startswith(self.www_root):
                    # Prevent escaping the www_root
                    warnings.warn(PossibleBreakout(entry.path, fspath))
                    continue
            is_dir = entry.is_dir()
            if is_dir:
                node_type = 'directory'
                slug = name
            elif self.is_dynamic(name):
                node_type = 'dynamic'
                slug = name.rsplit('.', 1)[0]
            else:
                node_type = 'static'
                slug = name
            if slug.startswith('%') and node_type != 'static':
                varname, vartype, extension = self.split_wildcard(slug[1:], is_dir)
                if varname in varnames:
                    raise WildcardCollision(varname, fspath)
                wildcard = '.'.join((varname, vartype)) if vartype else varname
                if is_dir:
                    slug = self.DIR_WILDCARD
                else:
                    node = FileNode(fspath, node_type, wildcard, extension)
                    wildleafs = children.setdefault(self.LEAF_WILDCARDS, {})
                    wildleafs[extension] = node
                    continue
            else:
                wildcard, extension = None, None
            if is_dir:
                if wildcard:
                    subvarnames = varnames.copy()
                    subvarnames[varname] = fspath
                else:
                    subvarnames = varnames
                subtree, mtime = self._build_subtree(fspath, subvarnames)
                node = self.make_dir_node(fspath, wildcard, subtree, mtime)
            else:
                node = FileNode(fspath, node_type, wildcard, extension)
            if slug in children:
                previous = children[slug]
                action = self.collision_handler(slug, previous, node)
                debug("collision: %r is claimed by both %r and %r | action: %r"
                     , slug, previous.fspath, node.fspath, action)
                if action == 'raise':
                    raise SlugCollision(slug, previous, node)
                if action == 'ignore_second_node':
                    continue
                if action == 'set_second_node_as_index_of_first_node':
                    previous.children[''] = node
                    continue
                if action != 'replace_first_node':
                    raise ValueError("%r is not a valid collision action" % action)
            children[slug] = node
            if fspath == index:
                children[''] = node
        return children, mtime

    def dispatch(self, path, path_segments):
        """"""
        DIR_WILDCARD = self.DIR_WILDCARD
        LEAF_WILDCARDS = self.LEAF_WILDCARDS

        extension, canonical = None, None
        fallback_wildleafs = None

        def fallback():
            if fallback_wildleafs:
                f_wildleafs, f_depth = fallback_wildleafs
                requested_extension = splitext(path_segments[-1])[1]
                if requested_extension in f_wildleafs:
                    node = f_wildleafs[requested_extension]
                elif None in f_wildleafs:
                    node = f_wildleafs[None]
                else:
                    debug("no suitable wildleaf fallback")
                    return MISSING
                debug("falling back to wild leaf: %r", node)
                if wildcards and f_depth < depth:
                    # We need to recreate the wildcards dict from scratch.
                    wildcards.clear()
                    fspath_segments = node.fspath[len(self.www_root)+1:].split(os.path.sep)
                    for i in range(f_depth):
                        fspath_segment = fspath_segments[i]
                        if fspath_segment.startswith('%'):
                            wildcards[fspath_segment[1:]] = path_segments[i]
                tail = '/'.join(path_segments[f_depth:])
                if node.extension:
                    wildcards[node.wildcard] = tail[:-len(node.extension)-1]
                else:
                    wildcards[node.wildcard] = tail
                return DispatchResult(DispatchStatus.okay, node.fspath, wildcards, None, None)
            debug("no wildleaf fallback")
            return MISSING

        def success():
            return DispatchResult(
                DispatchStatus.okay, node.fspath, wildcards, extension, canonical
            )

        wildcards = {}
        node = self.tree
        max_depth = len(path_segments) - 1
        for depth, segment in enumerate(path_segments):
            children = node.children

            if segment:
                # The segment isn't empty. Look for a match in children.
                if segment in children:
                    node = children[segment]
                    debug("exact match: %r -> %r", segment, node)
                    if node.type == 'directory':
                        continue
                    if depth == max_depth:
                        if node is children.get(''):
                            # The canonical path of `/index.html` is `/`
                            canonical = path[:-len(segment)]
                        return success()
                    elif depth == max_depth - 1 and path_segments[-1] == '':
                        # There is an extra slash at the end of the URL path.
                        if '.' not in segment:
                            canonical = path[:-1]
                            return success()
                if depth == max_depth and '.' in segment:
                    base, extension = segment.rsplit('.', 1)
                    if base in children and children[base].type == 'dynamic':
                        # Base match (e.g. `foo.spt` for `/foo.json`)
                        node = children[base]
                        debug("base match: %r -> %r", base, node)
                        if segment == node.fspath.rsplit(os.path.sep, 1)[-1]:
                            # Don't route a request for `/bar.html.spt` to `bar.html.spt`
                            return MISSING
                        return success()
                    extension = None
            elif depth == max_depth:
                # The segment is empty and it's the last one.
                break

            # This segment hasn't matched anything so far, look for wildcards.
            if LEAF_WILDCARDS in children:
                fallback_wildleafs = (children[LEAF_WILDCARDS], depth)
                debug("found fallback wildleafs: %r", fallback_wildleafs)
            if DIR_WILDCARD in children:
                # Try to find a wildleaf match first, so that `/foo.txt` matches
                # `/%bar.txt.spt` instead of `/%dir/`.
                if fallback_wildleafs and depth == max_depth:
                    result = fallback()
                    if result.status == DispatchStatus.okay:
                        return result
                # No suitable wildleaf was found, match the virtual directory.
                # Note: empty segments are allowed on purpose.
                node = children[DIR_WILDCARD]
                debug("virtual directory match: %r", node.wildcard)
                wildcards[node.wildcard] = segment
                continue

            return fallback()

        if node.type == 'directory':
            debug("final node is a directory")
            children = node.children
            canonical = path + '/' if path_segments[-1] != '' else None
            # Look for an index file
            if '' in children:
                debug("index match")
                node = children['']
            elif LEAF_WILDCARDS in children:
                debug("wildleaf match")
                wildleafs = children[LEAF_WILDCARDS]
                # Legacy behavior: dispatch to the "first" wildleaf
                node = wildleafs[min(wildleafs)]
                wildcards[node.wildcard] = ''
            else:
                # e.g. request for `/bar` is matched to empty wildcard directory `%foo/`
                result = fallback()
                if result.status == DispatchStatus.okay:
                    return result
                fspath = node.fspath + os.path.sep
                return DispatchResult(
                    DispatchStatus.unindexed, fspath, wildcards, extension, canonical
                )

        return success()


class HybridDispatcher(UserlandDispatcher):
    """A dispatcher optimized for development environments.

    This dispatcher is almost identical to :class:`UserlandDispatcher`, except
    that it does make some system calls to check that the matched filesystem
    directories haven't been modified. If changes are detected, then the
    dispacth tree is updated accordingly.

    This is the default dispatcher when the ``changes_reload`` configuration
    option is set to ``True``.
    """

    def make_dir_node(self, fspath, wildcard, children, mtime):
        return LiveDirectoryNode(fspath, wildcard, children, mtime, self)


class TestDispatcher(object):
    """
    This pseudo-dispatcher calls all the other dispatchers and checks that their
    results are identical. It's only meant to be used in Aspen's own tests.
    """

    def __init__(self, *args, **kw):
        self.dispatchers = [cls(*args, **kw) for cls in DISPATCHER_CLASSES]

    def build_dispatch_tree(self):
        for dispatcher in self.dispatchers:
            dispatcher.build_dispatch_tree()

    def dispatch(self, path, path_segments):
        results = [
            (dispatcher, dispatcher.dispatch(path, path_segments))
            for dispatcher in self.dispatchers
        ]
        if len(set(t[1]._as_tuple() for t in results)) != 1:
            raise AssertionError(
                "the dispatchers disagree:\n    " +
                "\n    ".join(
                    "%s returned %r" % (dispatcher.__class__.__name__, result)
                    for dispatcher, result in results
                )
            )
        return results[0][1]


DISPATCHER_CLASSES = [
    o for o in globals().values() if isclass(o) and issubclass(o, Dispatcher) and o != Dispatcher
]
DISPATCHER_CLASSES.sort(key=lambda c: c.__name__)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from aspen.simplates import pagination

#SPLIT TESTS
############

def check_page_content(raw, comp_pages):
    '''
    Pattern function. Splits a raw, then checks the number of pages generated,
    and that each page's content matches the contents of comp_pages.
    Interpretation of comp_pages is as follows:
    comp_pages is am item or a list of items. Each item is a string, tuple, or
    None. If it is a string, the page's content is matched; if it is a tuple,
    the page's content and/or header are checked. If any of the items are None,
    that comparison is ignored.
    '''

    # Convert to single-element list
    if not isinstance(comp_pages, list):
        comp_pages = [comp_pages]

    # Convert all non-tuples to tuples
    comp_pages = [item if isinstance(item, tuple) else (item, None)
                  for item in comp_pages]

    # Split into pages
    pages = list(pagination.split(raw))

    assert len(pages) == len(comp_pages)

    for generated_page, (content, header) in zip(pages, comp_pages):
        if content is not None:
            assert generated_page.content == content, repr(generated_page.content) + " should be " + repr(content)
        if header is not None:
            assert generated_page.header == header, repr(generated_page.header) + " should be " + repr(header)

def test_empty_content():
    check_page_content('', '')

def test_no_page_breaks():
    content = 'this is some content\nwith newlines'
    check_page_content(content, content)

def test_only_page_break():
    check_page_content('[---]\n', ['', ''])

def test_basic_page_break():
    check_page_content('Page 1\n[---]\nPage 2\n',
                       ['Page 1\n', 'Page 2\n'])

def test_two_page_breaks():
    raw = '''\
1
[---]
2
[---]
3
'''
    check_page_content(raw, ['1\n', '2\n', '3\n'])

def test_no_inline_page_break():
    content = 'this is an[---]inline page break'
    check_page_content(content,  [None])

def test_headers():
    raw = '''\
page1
[---] header2
page2
[---] header3
page3
'''
    pages = [
        ('page1\n', ''),
        ('page2\n', 'header2'),
        ('page3\n', 'header3')]

    check_page_content(raw, pages)
#ESCAPE TESTS
#############

def check_escape(content_to_escape, expected):
    actual = pagination.escape(content_to_escape)
    assert actual == expected, repr(actual) + " should be " + repr(expected)

def test_basic_escape_1():
    check_escape(r'\[---]', '[---]')

def test_basic_escape_2():
    check_escape(r'\\\[---]', r'\\[---]')

def test_inline_sep_ignored_1():
    check_escape('inline[---]break', 'inline[---]break')

def test_inline_sep_ignored_2():
    check_escape(r'inline\\[---]break', r'inline\\[---]break')

def test_escape_preserves_extra_content():
    check_escape(r'\\[---] content ', r'\[---] content ')

def test_multiple_escapes():
    to_escape = '1\n\\[---]\n2\n\\[---]'
    result = '1\n[---]\n2\n[---]'
    check_escape(to_escape, result)

def test_long_break():
    check_escape(r'\[----------]', '[----------]')

def test_escaped_pages():
    raw = '''\
1
[---]
2
\\[---]
3
'''
    check_page_content(raw, ['1\n', '2\n\\[---]\n3\n'])

#SPECLINE TESTS
###############

def check_specline(header, media_type, renderer):
    assert pagination.parse_specline(header) == (media_type, renderer)

def test_empty_header_1():
    check_specline('', '', '')

def test_empty_header_2():
    check_specline('    ', '', '')

def test_media_only():
    check_specline('text/plain', 'text/plain', '')

def test_renderer_only():
    check_specline('via renderer', '', 'renderer')

def test_basic_specline():
    check_specline('media/type via renderer', 'media/type', 'renderer')

def test_funky_whitespace():
    check_specline( '  media/type    via   renderer  '
                  , 'media/type'
                  , 'renderer'
                   )

def test_whitespace_in_fields():
    check_specline( 'media type via content renderer'
                  , 'media type'
                  , 'content renderer'
                   )

def test_extra_funky_whitespace():
    header = '   this is a  type   via some sort   of renderer    '
    media_type = 'this is a  type'
    renderer = 'some sort   of renderer'
    check_specline(header, media_type, renderer)



import os.path

import mimeparse
import mimetypes

from ..exceptions import AttemptedBreakout, NegotiationFailure, NotFound
from ..output import Output


def _is_subpath(path, root):
    return path.startswith(root) and path[len(root)] == os.path.sep


def open_resource(request_processor, resource_path):
    """Open a resource in read-only binary mode, after checking for symlinks.

    :raises AttemptedBreakout:
        if the :obj:`resource_path` points to a file that isn't inside any of
        the known
        :attr:`~aspen.request_processor.DefaultConfiguration.resource_directories`

    This function doesn't fully protect against attackers who have the ability
    to create and delete symlinks inside the resource directories whenever they
    want, but it makes the attack more difficult and detectable.
    """
    real_path = os.path.realpath(resource_path)
    is_outside = all(
        not _is_subpath(real_path, resource_dir)
        for resource_dir in request_processor.resource_directories
    )
    if is_outside:
        raise AttemptedBreakout(resource_path, real_path)
    return open(real_path, 'rb')


class Static(object):
    """Model a static HTTP resource.
    """

    __slots__ = ('request_processor', 'fspath', 'raw', 'media_type', 'charset')

    def __init__(self, request_processor, fspath):
        raw = None
        read_file = (
            request_processor.store_static_files_in_ram or
            request_processor.charset_static
        )
        if read_file:
            with open_resource(request_processor, fspath) as f:
                raw = f.read()
        self.request_processor = request_processor
        self.fspath = fspath
        self.raw = raw if request_processor.store_static_files_in_ram else None
        self.media_type = request_processor.guess_media_type(fspath)
        self.charset = None
        if request_processor.charset_static:
            try:
                raw.decode(request_processor.charset_static)
                self.charset = request_processor.charset_static
            except UnicodeDecodeError:
                pass

    def render(self, *ignored):
        """Returns the file's content as :class:`bytes`.

        If the ``store_static_files_in_ram`` configuration option was set to
        :obj:`False` (the default), then the file is read from the filesystem,
        otherwise its content is returned directly.
        """
        output = Output(media_type=self.media_type, charset=self.charset)
        if self.raw is None:
            with open_resource(self.request_processor, self.fspath) as f:
                output.body = f.read()
        else:
            output.body = self.raw
        return output


class Dynamic(object):
    """Model a dynamic HTTP resource.
    """

    __slots__ = ('request_processor', 'available_types')

    def render(self, context, dispatch_result, accept_header):
        """Render the resource.

        Before rendering we need to determine what type of content we're going
        to send back, by trying to find a match between the media types the
        client wants and the ones provided by the resource.

        The two sources for what the client wants are the extension in the
        request URL (:obj:`dispatch_result.extension`), and the ``Accept``
        header (:obj:`accept_header`). If the former fails to match we raise
        :class:`NotFound` (404), if the latter fails we raise
        :class:`NegotiationFailure` (406).

        Note that we don't always respect the ``Accept`` header (the spec says
        we can ignore it: <https://tools.ietf.org/html/rfc7231#section-5.3.2>).

        Args:
            context (dict): the variables you want to pass to the resource
            dispatch_result (DispatchResult): the object returned by the dispatcher
            accept_header (str): the requested media types

        Returns: an :class:`Output` object.

        """
        available = self.available_types

        dispatch_extension = dispatch_result.extension
        if dispatch_extension:
            # There's an extension in the URI path, guess the media type from it
            dispatch_accept = mimetypes.guess_type('a.' + dispatch_extension, strict=False)[0]
            if dispatch_accept is None:
                # The extension is unknown, raise NotFound
                raise NotFound()
            accept = dispatch_accept
            # Accept `media/type` for `media/x-type`
            i = accept.find('/x-')
            if i > 0:
                accept += ',' + accept[:i+1] + accept[i+3:]
            # Accept custom JSON media type
            if accept == 'application/json':
                accept += ',' + self.request_processor.media_type_json
        elif len(available) == 1:
            # If there's only one available type and no extension in the path,
            # then we ignore the Accept header
            return self.render_for_type(available[0], context)
        else:
            dispatch_accept = None
            accept = accept_header

        if accept:
            try:
                best_match = mimeparse.best_match(available, accept)
            except ValueError:
                # Unparseable accept header
                best_match = None
            if best_match:
                return self.render_for_type(best_match, context)
            elif best_match == '':
                if dispatch_accept is not None:
                    # e.g. client requested `/foo.json` but `/foo.spt` has no JSON page
                    raise NotFound()
                raise NegotiationFailure(accept, available)

        # Fall back to the first available type
        return self.render_for_type(available[0], context)



# for backwards compatibility with aspen-renderer modules
from .simplates.renderers import Factory, Renderer

Factory, Renderer # make pyflakes happy

import warnings
warnings.warn('aspen.renderers is deprecated and will be removed in a future version. '
              'Please use aspen.simplates.renderers instead.', FutureWarning)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import pytest
from pytest import raises, fixture

from aspen.exceptions import NegotiationFailure, NotFound
from aspen.http.resource import mimetypes
from aspen.simplates.simplate import _decode
from aspen.simplates.simplate import Simplate
from aspen.simplates.pagination import Page
from aspen.simplates.renderers.stdlib_template import Factory as TemplateFactory
from aspen.simplates.renderers.stdlib_percent import Factory as PercentFactory


@fixture
def make_simplate(harness):
    def get(fspath='index.spt', raw=b'[---]\n[---] text/plain via stdlib_template\n'):
        harness.fs.www.mk((fspath, raw, False))
        return Simplate(harness.request_processor, harness.fs.www.resolve(fspath))
    yield get


def test_dynamic_resource_is_instantiable(make_simplate):
    actual = make_simplate().__class__
    assert actual is Simplate


def test_duplicate_media_type_causes_SyntaxError(make_simplate):
    with raises(SyntaxError):
        make_simplate(raw=b"[---]\n[---] text/plain\n[---] text/plain\n")


# compile_page

def test_compile_page_compiles_empty_page(make_simplate):
    page = make_simplate().compile_page(Page('', 'text/html'))
    actual = page[0]({}), page[1]
    assert actual == ('', 'text/html')

def test_compile_page_compiles_page(make_simplate):
    page = make_simplate().compile_page(Page('foo bar', 'text/html'))
    actual = page[0]({}), page[1]
    assert actual == ('foo bar', 'text/html')


# _parse_specline

def test_parse_specline_parses_specline(make_simplate):
    factory, media_type = make_simplate()._parse_specline('media/type via stdlib_template')
    actual = (factory.__class__, media_type)
    assert actual == (TemplateFactory, 'media/type')

def test_parse_specline_doesnt_require_renderer(make_simplate):
    factory, media_type = make_simplate()._parse_specline('media/type')
    actual = (factory.__class__, media_type)
    assert actual == (PercentFactory, 'media/type')

def test_parse_specline_doesnt_require_media_type(harness, make_simplate):
    factory, media_type = make_simplate()._parse_specline('via stdlib_template')
    actual = (factory.__class__, media_type)
    assert actual == (TemplateFactory, harness.request_processor.media_type_default)

def test_parse_specline_raises_SyntaxError_if_renderer_is_malformed(make_simplate):
    with raises(SyntaxError):
        make_simplate()._parse_specline('stdlib_template media/type')

def test_parse_specline_raises_SyntaxError_if_media_type_is_malformed(make_simplate):
    with raises(SyntaxError):
        make_simplate()._parse_specline('media-type via stdlib_template')

def test_parse_specline_cant_mistake_malformed_media_type_for_renderer(make_simplate):
    with raises(SyntaxError):
        make_simplate()._parse_specline('media-type')

def test_parse_specline_cant_mistake_malformed_renderer_for_media_type(make_simplate):
    with raises(SyntaxError):
        make_simplate()._parse_specline('stdlib_template')

def test_parse_specline_enforces_order(make_simplate):
    with raises(SyntaxError):
        make_simplate()._parse_specline('stdlib_template via media/type')

def test_parse_specline_obeys_default_by_media_type(make_simplate):
    resource = make_simplate()
    resource.default_renderers_by_media_type['media/type'] = 'glubber'
    err = raises(ValueError, resource._parse_specline, 'media/type').value
    msg = err.args[0]
    assert msg.startswith("Unknown renderer for media/type: glubber."), msg

def test_parse_specline_obeys_default_by_media_type_default(make_simplate):
    resource = make_simplate()
    resource.default_renderers_by_media_type.default_factory = lambda: 'glubber'
    err = raises(ValueError, resource._parse_specline, 'media/type').value
    msg = err.args[0]
    assert msg.startswith("Unknown renderer for media/type: glubber.")

def test_get_renderer_factory_can_raise_syntax_error(make_simplate):
    resource = make_simplate()
    resource.default_renderers_by_media_type['media/type'] = 'glubber'
    err = raises( SyntaxError
                       , resource._get_renderer_factory
                       , 'media/type'
                       , 'oo*gle'
                        ).value
    msg = err.args[0]
    assert msg.startswith("Malformed renderer oo*gle. It must match")


# render

SIMPLATE = """\
[---]
[---] text/plain
Greetings, program!
[---] text/html
<h1>Greetings, program!</h1>
"""

def test_render_is_happy_not_to_negotiate(harness):
    output = harness.simple(filepath='index.spt', contents=SIMPLATE)
    assert output.text == "Greetings, program!\n"

def test_render_sets_media_type_when_it_doesnt_negotiate(harness):
    output = harness.simple(filepath='index.spt', contents=SIMPLATE)
    assert output.media_type == "text/plain"

def test_render_is_happy_not_to_negotiate_with_defaults(harness):
    output = harness.simple(filepath='index.spt', contents="[---]\n[---]\nGreetings, program!\n")
    assert output.text == "Greetings, program!\n"

def test_render_negotiates(harness):
    output = harness.simple(filepath='index.spt', contents=SIMPLATE, accept_header='text/html')
    assert output.text == "<h1>Greetings, program!</h1>\n"

def test_ignores_busted_accept(harness):
    output = harness.simple(filepath='index.spt', contents=SIMPLATE, accept_header='text/html/foo')
    assert output.text == "Greetings, program!\n"

def test_render_sets_media_type_when_it_negotiates(harness):
    output = harness.simple(filepath='index.spt', contents=SIMPLATE, accept_header='text/html')
    assert output.media_type == "text/html"

def test_render_raises_if_direct_negotiation_fails(harness):
    with raises(NegotiationFailure):
        harness.simple(filepath='index.spt', contents=SIMPLATE, accept_header='cheese/head')

def test_render_negotation_failures_include_available_types(harness):
    actual = raises(
        NegotiationFailure,
        harness.simple, filepath='index.spt', contents=SIMPLATE, accept_header='cheese/head'
    ).value.message
    expected = "Couldn't satisfy cheese/head. The following media types are available: " \
               "text/plain, text/html."
    assert actual == expected

def test_treat_media_type_variants_as_equivalent(harness):
    _guess_type = mimetypes.guess_type
    mimetypes.guess_type = lambda url, **kw: ('application/x-javascript' if url.endswith('.js') else '', None)
    try:
        output = harness.simple(
            filepath='foobar.spt',
            contents="[---]\n[---] application/javascript\n[---] text/plain\n",
            uripath='/foobar.js',
        )
        assert output.media_type == "application/javascript"
    finally:
        mimetypes.guess_type = _guess_type


from aspen.simplates.renderers import Renderer, Factory

class Glubber(Renderer):
    def render_content(self, context):
        return "glubber"

class GlubberFactory(Factory):
    Renderer = Glubber

class glubber:

    def __init__(self, harness):
        self.harness = harness

    def __enter__(self):
        Simplate.renderer_factories['glubber'] = GlubberFactory(self.harness.request_processor)
        self.__old = Simplate.default_renderers_by_media_type['text/plain']
        Simplate.default_renderers_by_media_type['text/plain'] = 'glubber'

    def __exit__(self, *a, **kw):
        del Simplate.renderer_factories['glubber']
        Simplate.default_renderers_by_media_type['text/plain'] = self.__old


def test_can_override_default_renderers_by_mimetype(harness):
    with glubber(harness):
        harness.fs.www.mk(('index.spt', SIMPLATE),)
        output = harness.simple(filepath='index.spt', contents=SIMPLATE, accept_header='text/plain')
        assert output.text == "glubber"

def test_can_override_default_renderer_entirely(harness):
    with glubber(harness):
        output = harness.simple(filepath='index.spt', contents=SIMPLATE, accept_header='text/plain')
        assert output.text == "glubber"


# indirect

INDIRECTLY_NEGOTIATED_SIMPLATE = """\
[-------]
foo = "program"
[-------] text/html
<h1>Greetings, %(foo)s!</h1>
[-------] text/plain
Greetings, %(foo)s!"""

def test_indirect_negotiation_sets_media_type(harness):
    output = harness.simple(INDIRECTLY_NEGOTIATED_SIMPLATE, '/foo.spt', '/foo.html')
    expected = "<h1>Greetings, program!</h1>\n"
    actual = output.text
    assert actual == expected

def test_indirect_negotiation_sets_media_type_to_secondary(harness):
    output = harness.simple(INDIRECTLY_NEGOTIATED_SIMPLATE, '/foo.spt', '/foo.txt')
    expected = "Greetings, program!"
    actual = output.text
    assert actual == expected

def test_indirect_negotiation_with_unsupported_media_type_is_an_error(harness):
    with raises(NotFound):
        harness.simple(INDIRECTLY_NEGOTIATED_SIMPLATE, '/foo.spt', '/foo.jpg')


SIMPLATE_VIRTUAL_PATH = """\
[-------]
foo = path['foo']
[-------] text/html
<h1>Greetings, %(foo)s!</h1>
[-------] text/plain
Greetings, %(foo)s!"""


def test_dynamic_resource_inside_virtual_path(harness):
    output = harness.simple(SIMPLATE_VIRTUAL_PATH, '/%foo/bar.spt', '/program/bar.txt')
    expected = "Greetings, program!"
    actual = output.text
    assert actual == expected

SIMPLATE_STARTYPE = """\
[-------]
foo = path['foo']
[-------] */*
Unknown request type, %(foo)s!
[-------] text/html
<h1>Greetings, %(foo)s!</h1>
[-------] text/*
Greetings, %(foo)s!"""

def test_dynamic_resource_inside_virtual_path_with_startypes_present(harness):
    output = harness.simple(SIMPLATE_STARTYPE, '/%foo/bar.spt', '/program/bar.html')
    actual = output.text
    assert '<h1>' in actual

def test_dynamic_resource_inside_virtual_path_with_startype_partial_match(harness):
    output = harness.simple(SIMPLATE_STARTYPE, '/%foo/bar.spt', '/program/bar.txt')
    expected = "Greetings, program!"
    actual = output.text
    assert actual == expected

def test_dynamic_resource_inside_virtual_path_with_startype_fallback(harness):
    output = harness.simple(SIMPLATE_STARTYPE, '/%foo/bar.spt', '/program/bar.jpg')
    expected = "Unknown request type, program!"
    actual = output.text.strip()
    assert actual == expected


def test_default_media_type_works(harness):
    output = harness.simple("""
[---]
[---]
plaintext""")
    assert "plaintext" in output.text


def test_but_python_sections_exhibit_module_scoping_behavior(harness):
    output = harness.simple("""[---]
bar = 'baz'
def foo():
    return bar
foo = foo()
[---] text/html via stdlib_format
{foo}""")
    assert output.text == 'baz'


def test_one_page_works(harness):
    output = harness.simple("Template")
    assert output.text == 'Template'


def test_two_pages_works(harness):
    output = harness.simple("""
foo = 'Template'
[---] via stdlib_format
{foo}""")
    assert output.text == 'Template'


def test_three_pages_one_python_works(harness):
    output = harness.simple("""
foo = 'Template'
[---] text/plain via stdlib_format
{foo}
[---] text/xml
<foo>{foo}</foo>""", filepath='index.spt')
    assert output.text.strip() == 'Template'


def test_three_pages_two_python_works(harness):
    output = harness.simple("""[---]
python_code = True
[---]
Template""")
    assert output.text == 'Template'


# _decode

def test_decode_can_take_encoding_from_first_line():
    actual = _decode("""\
    # -*- coding: utf8 -*-
    text = u'א'
    """.encode('utf8'))
    expected = """\
    # encoding set to utf8
    text = u'א'
    """
    assert actual == expected

def test_decode_can_take_encoding_from_second_line():
    actual = _decode("""\
    #!/blah/blah
    # -*- coding: utf8 -*-
    text = u'א'
    """.encode('utf8'))
    expected = """\
    #!/blah/blah
    # encoding set to utf8
    text = u'א'
    """
    assert actual == expected

def test_decode_prefers_first_line_to_second():
    actual = _decode("""\
    # -*- coding: utf8 -*-
    # -*- coding: ascii -*-
    text = u'א'
    """.encode('utf8'))
    expected = """\
    # encoding set to utf8
    # encoding NOT set to ascii
    text = u'א'
    """
    assert actual == expected

def test_decode_ignores_third_line():
    actual = _decode("""\
    # -*- coding: utf8 -*-
    # -*- coding: ascii -*-
    # -*- coding: cornnuts -*-
    text = u'א'
    """.encode('utf8'))
    expected = """\
    # encoding set to utf8
    # encoding NOT set to ascii
    # -*- coding: cornnuts -*-
    text = u'א'
    """
    assert actual == expected

formats = [ '-*- coding: utf8 -*-'
          , '-*- encoding: utf8 -*-'
          , 'coding: utf8'
          , '  coding: utf8'
          , '\tencoding: utf8'
          , '\t flubcoding=utf8'
           ]
@pytest.mark.parametrize('fmt', formats)
def test_decode_can_take_encoding_from_various_line_formats(fmt):
    actual = _decode("""\
    # {0}
    text = u'א'
    """.format(fmt).encode('utf8'))
    expected = """\
    # encoding set to utf8
    text = u'א'
    """
    assert actual == expected

formats = [ '-*- coding : utf8 -*-'
          , 'foo = 0 -*- encoding: utf8 -*-'
          , '  coding : utf8'
          , 'encoding : utf8'
          , '  flubcoding =utf8'
          , 'coding: '
           ]
@pytest.mark.parametrize('fmt', formats)
def test_decode_cant_take_encoding_from_bad_line_formats(fmt):
    def test():
        raw = """\
        # {0}
        text = u'א'
        """.format(fmt).encode('utf8')
        raises(UnicodeDecodeError, _decode, raw)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import pytest

import aspen


def test_virtual_path_with_typecast(harness):
    harness.fs.www.mk(('%year.int/foo.html', "Greetings, program!"),)
    path = harness.hit('/1999/foo.html', want='path')
    assert path == {'year': [1999]}

def test_virtual_path_raises_on_bad_typecast(harness):
    harness.fs.www.mk(('%year.int/foo.html', "Greetings, program!"),)
    with pytest.raises(aspen.exceptions.TypecastError):
        harness.hit('/I am not a year./foo.html')

class User:

    def __init__(self, name):
        self.username = name

    @classmethod
    def toUser(cls, name, context):
        return cls(name)

def test_virtual_path_file_key_val_cast_custom(harness):
    harness.fs.www.mk(('user/%user.user.html.spt', "Greetings, user!"))
    harness.hydrate_request_processor(typecasters={'user': User.toUser})
    actual = harness.hit('/user/chad.html', want='path')
    assert actual['user'].username == 'chad'



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from six import text_type as str

from pytest import raises

from aspen.simplates import json_
from aspen.simplates.renderers import Factory, Renderer
from aspen.simplates.simplate import Simplate


def test_a_custom_renderer(harness):
    class TestRenderer(Renderer):

        def compile(self, *a):
            return self.raw.upper()

        def render_content(self, context):
            d = dict((k, v) for k, v in self.__dict__.items() if k[0] != '_')
            return str(json_.dumps(d))

    class TestFactory(Factory):
        Renderer = TestRenderer

        def compile_meta(self, configuration):
            return 'foobar'

    request_processor = harness.request_processor
    try:
        Simplate.renderer_factories['lorem'] = TestFactory(request_processor)

        r = harness.simple("[---]\n[---] text/html via lorem\nLorem ipsum")
        assert r.text
        d = json_.loads(r.text)
        assert d['meta'] == 'foobar'
        assert d['raw'] == 'Lorem ipsum'
        assert d['media_type'] == 'text/html'
        assert d['offset'] == 2
        assert d['compiled'] == 'LOREM IPSUM'
    finally:
        del Simplate.renderer_factories['lorem']


def test_renderer_padding_works_with_padded_output(harness):
    class TestRenderer(Renderer):

        def compile(self, filepath, padded):
            assert padded[:self.offset] == '\n' * self.offset
            return padded

        def render_content(self, context):
            return '\n' + self.compiled + '\n'

    class TestFactory(Factory):
        Renderer = TestRenderer

    request_processor = harness.request_processor
    try:
        Simplate.renderer_factories['x'] = TestFactory(request_processor)

        output = harness.simple("[---]\n[---] text/plain via x\nSome text")
        assert output.text == '\nSome text\n'
    finally:
        del Simplate.renderer_factories['x']


def test_renderer_padding_works_with_stripped_output(harness):
    class TestRenderer(Renderer):

        def compile(self, filepath, padded):
            assert padded[:self.offset] == '\n' * self.offset
            return padded

        def render_content(self, context):
            return '\n' + self.raw + '\n'

    class TestFactory(Factory):
        Renderer = TestRenderer

    request_processor = harness.request_processor
    try:
        Simplate.renderer_factories['y'] = TestFactory(request_processor)

        output = harness.simple("[---]\n[---] text/plain via y\nSome text")
        assert output.text == '\nSome text\n'
    finally:
        del Simplate.renderer_factories['y']



def test_renderer_padding_achieves_correct_line_numbers_in_tracebacks(harness):
    harness.fs.www.mk(('index.html.spt', '''\
    [---]
    [---] text/html via stdlib_template
    Greetings, $!
    '''))
    actual = str(raises(ValueError, harness.hit, '/').value)
    assert 'line 3' in actual


def test_aspen_renderers_is_deprecated(harness):
    import warnings
    with warnings.catch_warnings(record=True) as w:
        assert w == []
        from aspen import renderers; renderers  # (; for pyflakes)
        assert [x.category for x in w] == [FutureWarning]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import pytest

from filesystem_tree import FilesystemTree

from aspen.exceptions import SlugCollision, WildcardCollision
from aspen.http.request import Path
from aspen.request_processor.dispatcher import (
    DISPATCHER_CLASSES, DispatchStatus, legacy_collision_handler
)


# Helpers
# =======

def assert_match(
    harness, request_path, expected_match,
    status=DispatchStatus.okay, wildcards=None, extension=None, canonical=None
):
    result = harness.request_processor.dispatch(Path(request_path))
    assert result.status == status
    assert result.match == harness.fs.www.resolve(expected_match)
    assert result.wildcards == wildcards
    assert result.extension == extension
    assert result.canonical == canonical

def assert_missing(harness, request_path):
    result = harness.request_processor.dispatch(Path(request_path))
    assert result.status == DispatchStatus.missing
    assert result.match is None

def assert_unindexed(harness, request_path, wildcards=None, extension=None, canonical=None):
    result = harness.request_processor.dispatch(Path(request_path))
    assert result.status == DispatchStatus.unindexed
    assert result.match == harness.fs.www.resolve(request_path) + os.path.sep
    assert result.wildcards == wildcards
    assert result.extension == extension
    assert result.canonical == canonical

def assert_wildcards(harness, request_path, expected_vals):
    result = harness.request_processor.dispatch(Path(request_path))
    assert result.wildcards == expected_vals

def dispatch(harness, request_path):
    return harness.request_processor.dispatch(Path(request_path))

NEGOTIATED_SIMPLATE="""[-----]
[-----] text/plain
Greetings, program!
[-----] text/html
<h1>Greetings, Program!</h1>"""


# dispatcher.dispatch
# ===================

@pytest.mark.parametrize('dispatcher_class', DISPATCHER_CLASSES)
def test_dispatcher_returns_a_result(dispatcher_class):
    www = FilesystemTree()
    www.mk(('index.html', 'Greetings, program!'),)
    dispatcher = dispatcher_class(
        www_root    = www.root,
        is_dynamic  = lambda n: n.endswith('.spt'),
        indices     = ['index.html'],
        typecasters = {},
    )
    dispatcher.build_dispatch_tree()
    result = dispatcher.dispatch('/', [''])
    assert result.status == DispatchStatus.okay
    assert result.match == os.path.join(www.root, 'index.html')
    assert result.wildcards == None

@pytest.mark.parametrize('dispatcher_class', DISPATCHER_CLASSES)
def test_dispatcher_returns_unindexed_for_unindexed_directory(dispatcher_class):
    www = FilesystemTree()
    dispatcher = dispatcher_class(
        www_root    = www.root,
        is_dynamic  = lambda n: n.endswith('.spt'),
        indices     = [],
        typecasters = {},
    )
    dispatcher.build_dispatch_tree()
    r = dispatcher.dispatch('/', [''])
    assert r.status == DispatchStatus.unindexed
    assert r.match == www.root + os.path.sep

def test_dispatch_when_filesystem_has_been_modified():
    # Create an empty www_root
    www = FilesystemTree()
    # Help the dispatchers that rely on `mtime`
    st = os.stat(www.root)
    os.utime(www.root, (st.st_atime - 10, st.st_mtime - 10))
    # Initialize the dispatchers
    dispatchers = []
    for dispatcher_class in DISPATCHER_CLASSES:
        dispatchers.append(dispatcher_class(
            www_root    = www.root,
            is_dynamic  = lambda n: n.endswith('.spt'),
            indices     = ['index.html'],
            typecasters = {},
        ))
        dispatchers[-1].build_dispatch_tree()
    # Now add an index file and try to dispatch
    www.mk(('index.html', 'Greetings, program!'))
    for dispatcher in dispatchers:
        print("Attempting dispatch with", dispatcher.__class__.__name__)
        result = dispatcher.dispatch('/', [''])
        if dispatcher.__class__.__name__ == 'UserlandDispatcher':
            assert result.status == DispatchStatus.unindexed
            assert result.match == www.root + os.path.sep
        else:
            assert result.status == DispatchStatus.okay
            assert result.match == www.resolve('index.html')
        assert result.wildcards is None
        assert result.extension is None
        assert result.canonical is None


# Indices
# =======

def test_static_index(harness):
    harness.fs.www.mk(('index.html', 'Greetings, program!'))
    assert_match(harness, '/', 'index.html')

def test_index_without_extention(harness):
    harness.fs.www.mk(('index', 'Greetings, program!'))
    assert_match(harness, '/', 'index')

def test_dynamic_index(harness):
    harness.fs.www.mk(('index.spt', NEGOTIATED_SIMPLATE))
    assert_match(harness, '/', 'index.spt')

def test_empty_root(harness):
    assert_unindexed(harness, '/')

def test_unrecognized_index(harness):
    harness.fs.www.mk(('index.html', "Greetings, program!"),)
    harness.hydrate_request_processor(indices=["default.html"])
    assert_unindexed(harness, '/')

def test_dispatcher_matches_second_index_if_first_is_missing(harness):
    harness.fs.www.mk(('default.html', "Greetings, program!"),)
    harness.hydrate_request_processor(indices=["index.html", "default.html"])
    assert_match(harness, '/', 'default.html')

def test_dispatcher_matches_explicit_index(harness):
    harness.fs.www.mk(('index.html', "Greetings, program!"),)
    harness.hydrate_request_processor(indices=["index.html", "default.html"])
    assert_match(harness, '/index.html', 'index.html', canonical='/')

def test_dispatcher_matches_explicit_second_index(harness):
    harness.fs.www.mk(('default.html', "Greetings, program!"),)
    harness.hydrate_request_processor(indices=["index.html", "default.html"])
    assert_match(harness, '/default.html', 'default.html', canonical='/')

def test_dispatch_with_two_indexes(harness):
    harness.fs.www.mk(
        ('default.html', "Greetings, program!"),
        ('index.html', "Greetings, program!")
    )
    harness.hydrate_request_processor(indices=["index.html", "default.html"])
    assert_match(harness, '/', 'index.html')
    assert_match(harness, '/index.html', 'index.html', canonical='/')
    assert_match(harness, '/default.html', 'default.html')


# Negotiated Fall-through
# =======================

def test_indirect_negotiation_can_passthrough_static(harness):
    harness.fs.www.mk(('foo.html', "Greetings, program!"),)
    assert_match(harness, 'foo.html', 'foo.html')

def test_indirect_negotiation_can_passthrough_renderered(harness):
    harness.fs.www.mk(('foo.html.spt', NEGOTIATED_SIMPLATE),)
    assert_match(harness, 'foo.html', 'foo.html.spt')

def test_indirect_negotiation_can_passthrough_negotiated(harness):
    harness.fs.www.mk(('foo', "Greetings, program!"),)
    assert_match(harness, 'foo', 'foo')

def test_indirect_negotiation_modifies_one_dot(harness):
    harness.fs.www.mk(('foo.spt', NEGOTIATED_SIMPLATE),)
    assert_match(harness, 'foo.html', 'foo.spt', extension='html')

def test_indirect_negotiation_skips_two_dots(harness):
    harness.fs.www.mk(('foo.bar.spt', NEGOTIATED_SIMPLATE),)
    assert_match(harness, 'foo.bar.html', 'foo.bar.spt', extension='html')

def test_indirect_negotiation_prefers_rendered(harness):
    harness.fs.www.mk( ('foo.html', "Greetings, program!")
          , ('foo', "blah blah blah")
           )
    assert_match(harness, 'foo.html', 'foo.html')

def test_indirect_negotiation_really_prefers_rendered(harness):
    harness.fs.www.mk( ('foo.html', "Greetings, program!")
          , ('foo.', "blah blah blah")
           )
    assert_match(harness, 'foo.html', 'foo.html')

def test_indirect_negotiation_really_prefers_rendered_2(harness):
    harness.fs.www.mk( ('foo.html', "Greetings, program!")
          , ('foo', "blah blah blah")
           )
    assert_match(harness, 'foo.html', 'foo.html')

def test_indirect_negotation_doesnt_do_dirs(harness):
    assert_missing(harness, 'foo.html')


# Virtual Paths
# =============

def test_virtual_path_can_passthrough(harness):
    harness.fs.www.mk(('foo.html', "Greetings, program!"),)
    assert_match(harness, 'foo.html', 'foo.html')

def test_unfound_virtual_path_passes_through(harness):
    harness.fs.www.mk(('%bar/foo.html', "Greetings, program!"),)
    assert_missing(harness, '/blah/flah.html')

def test_virtual_path_is_virtual(harness):
    harness.fs.www.mk(('%bar/foo.html', "Greetings, program!"),)
    assert_match(harness, '/blah/foo.html', '%bar/foo.html', wildcards={'bar': 'blah'})

def test_virtual_path_sets_path(harness):
    harness.fs.www.mk(('%bar/foo.spt', NEGOTIATED_SIMPLATE),)
    assert_wildcards(harness, '/blah/foo.html', {'bar': 'blah'} )

def test_virtual_path_sets_unicode_path(harness):
    harness.fs.www.mk(('%bar/foo.html', "Greetings, program!"),)
    assert_wildcards(harness, '/%E2%98%83/foo.html', {'bar': '\u2603'})

def test_virtual_path_with_typecast(harness):
    harness.fs.www.mk(('%year.int/foo.html', "Greetings, program!"),)
    assert_wildcards(harness, '/1999/foo.html', {'year.int': '1999'})

def test_virtual_path_raises_on_direct_access(harness):
    assert_missing(harness, '/%name/foo.html')

def test_virtual_path_raises_404_on_direct_access(harness):
    assert_missing(harness, '/%name/foo.html')

def test_virtual_path_directory(harness):
    harness.fs.www.mk(('%first/index.html', "Greetings, program!"),)
    assert_match(harness, '/foo/', '%first/index.html', wildcards={'first': 'foo'})

def test_virtual_path_file(harness):
    harness.fs.www.mk(('foo/%bar.html.spt', NEGOTIATED_SIMPLATE),)
    assert_match(harness, '/foo/blah.html', 'foo/%bar.html.spt', wildcards={'bar': 'blah'})

def test_virtual_path_file_only_last_part(harness):
    harness.fs.www.mk(('foo/%bar.html.spt', NEGOTIATED_SIMPLATE),)
    assert_match(harness, '/foo/blah/baz.html', 'foo/%bar.html.spt', wildcards={'bar': 'blah/baz'})

def test_virtual_path_file_only_last_part____no_really(harness):
    harness.fs.www.mk(('foo/%bar.html.spt', "Greetings, program!"),)
    assert_missing(harness, '/foo/blah.html/')

def test_virtual_path_file_key_val_set(harness):
    harness.fs.www.mk(('foo/%bar.html.spt', NEGOTIATED_SIMPLATE),)
    assert_wildcards(harness, '/foo/blah.html', {'bar': 'blah'})

def test_virtual_path_file_key_val_without_cast(harness):
    harness.fs.www.mk(('foo/%bar.html.spt', NEGOTIATED_SIMPLATE),)
    assert_wildcards(harness, '/foo/537.html', {'bar': '537'})

def test_virtual_path_file_key_val_with_cast(harness):
    harness.fs.www.mk(('foo/%bar.int.html.spt', NEGOTIATED_SIMPLATE),)
    assert_wildcards(harness, '/foo/537.html', {'bar.int': '537'})

def test_virtual_path_file_key_val_percent(harness):
    harness.fs.www.mk(('foo/%bar.spt', NEGOTIATED_SIMPLATE),)
    assert_wildcards(harness, '/foo/%25blah', {'bar': '%blah'})

def test_virtual_path_file_not_dir(harness):
    harness.fs.www.mk( ('%foo/bar.html', "Greetings from bar!")
          , ('%baz.html.spt', NEGOTIATED_SIMPLATE)
           )
    assert_match(harness, '/bal.html', '%baz.html.spt', wildcards={'baz': 'bal'})

def test_virtual_path_when_dir_is_more_specific(harness):
    harness.fs.www.mk(
        ('%foo/%bar/file.html.spt', "Hello world!"),
        ('%foo.html.spt', "Hello world!")
    )
    assert_wildcards(harness, '/~/x/file.html', {'foo': '~', 'bar': 'x'})

def test_static_files_are_not_wild(harness):
    harness.fs.www.mk(('foo/%bar.html', "Greetings, program!"),)
    assert_missing(harness, '/foo/blah.html')
    assert_match(harness, '/foo/%25bar.html', 'foo/%bar.html')

def test_fallback_to_wildleaf_in_parent_directory(harness):
    harness.fs.www.mk(
        ('%foo/%bar/file.html', "Hello world!"),
        ('%foo/%bar.html', "Hello world!"),
        ('%foo.spt', NEGOTIATED_SIMPLATE)
    )
    assert_match(harness, '/~/x/file.txt', '%foo.spt', wildcards={'foo': '~/x/file.txt'})


# negotiated *and* virtual paths
# ==============================

def test_virtual_path__and_indirect_neg_file_not_dir(harness):
    harness.fs.www.mk( ('%foo/bar.html', "Greetings from bar!")
          , ('%baz.spt', NEGOTIATED_SIMPLATE)
           )
    assert_match(harness, '/bal.html', '%baz.spt', wildcards={'baz': 'bal.html'})

def test_virtual_path_and_indirect_neg_noext(harness):
    harness.fs.www.mk(('%foo/bar', "Greetings program!"),)
    assert_match(harness, '/greet/bar', '%foo/bar', wildcards={'foo': 'greet'})

def test_virtual_path_and_indirect_neg_ext(harness):
    harness.fs.www.mk(('%foo/bar.spt', NEGOTIATED_SIMPLATE),)
    assert_match(harness, '/greet/bar.html', '%foo/bar.spt',
                 wildcards={'foo': 'greet'}, extension='html')


# Collisions
# ==========

def test_variable_name_used_twice_in_path_results_in_WildcardCollision(harness):
    harness.fs.www.mk(('%foo/bar/%foo.spt', "Hello world!"))
    with pytest.raises(WildcardCollision):
        harness.hydrate_request_processor()

def test_same_name_used_in_different_dirs_is_okay(harness):
    harness.fs.www.mk(
        ('foo/%bar.spt', "Hello world!"),
        ('foo/z/%bar.spt', "Hello world!"),
    )
    assert_match(harness, '/foo/y/bar', 'foo/%bar.spt', wildcards={'bar': 'y/bar'})
    assert_match(harness, '/foo/z/bar', 'foo/z/%bar.spt', wildcards={'bar': 'bar'})

def test_same_dir_name_used_in_different_paths_is_okay(harness):
    harness.fs.www.mk(
        ('%foo/bar.spt', "Hello world!"),
        ('x/%foo/bar.spt', "Hello world!"),
    )
    assert_match(harness, '/foo/bar', '%foo/bar.spt', wildcards={'foo': 'foo'})
    assert_match(harness, '/x/foo/bar', 'x/%foo/bar.spt', wildcards={'foo': 'foo'})

def test_collision_between_file_and_directory(harness):
    harness.fs.www.mk(
        ('foo/bar.spt', ''),
        ('foo/bar/index.spt', ''),
    )
    with pytest.raises(SlugCollision):
        harness.hydrate_request_processor()

def test_collision_between_two_dir_wildcards(harness):
    harness.fs.www.mk(
        ('%first/foo.html', "Greetings, program!"),
        ('%second/foo.html', "WWAAAAAAAAAAAA!!!!!!!!"),
    )
    # Test with default collision handler, should raise an exception.
    with pytest.raises(SlugCollision):
        harness.hydrate_request_processor()
    # Test with legacy collision handler, should favor the first node.
    harness.hydrate_request_processor(
        dispatcher_options=dict(collision_handler=legacy_collision_handler),
    )
    assert_match(harness, '/1999/foo.html', '%first/foo.html', wildcards={'first': '1999'})


# trailing slash
# ==============

def test_dispatcher_passes_through_files(harness):
    harness.fs.www.mk(('foo/index.html', "Greetings, program!"),)
    assert_missing(harness, '/foo/537.html')

def test_trailing_slash_passes_dirs_with_slash_through(harness):
    harness.fs.www.mk(('foo/index.html', "Greetings, program!"),)
    assert_match(harness, '/foo/', '/foo/index.html')

def test_dispatcher_passes_through_virtual_dir_with_trailing_slash(harness):
    harness.fs.www.mk(('%foo/index.html', "Greetings, program!"),)
    assert_match(harness, '/foo/', '/%foo/index.html', wildcards={'foo': 'foo'})

def test_dispatcher_matches_dir_even_without_trailing_slash(harness):
    harness.fs.www.mk('foo',)
    assert_unindexed(harness, '/foo', canonical='/foo/')

def test_dispatcher_matches_virtual_dir_even_without_trailing_slash(harness):
    harness.fs.www.mk('%foo',)
    result = dispatch(harness, '/foo')
    assert result.status == DispatchStatus.unindexed
    assert result.match == harness.fs.www.resolve('%foo') + os.path.sep
    assert result.wildcards == {'foo': 'foo'}
    assert result.canonical == '/foo/'

def test_trailing_on_virtual_paths_missing(harness):
    harness.fs.www.mk('%foo/%bar/%baz/',)
    result = dispatch(harness, '/foo/bar/baz')
    assert result.status == DispatchStatus.unindexed
    assert result.match == harness.fs.www.resolve('%foo/%bar/%baz') + os.path.sep
    assert result.wildcards == {'foo': 'foo', 'bar': 'bar', 'baz': 'baz'}
    assert result.canonical == '/foo/bar/baz/'

def test_trailing_on_virtual_paths(harness):
    harness.fs.www.mk(('%foo/%bar/%baz/index.html', "Greetings program!"),)
    assert_match(harness, '/foo/bar/baz/', '/%foo/%bar/%baz/index.html',
                 wildcards={'foo': 'foo', 'bar': 'bar', 'baz': 'baz'})

def test_trailing_slash_matches_wild_leaf(harness):
    harness.fs.www.mk(('foo/%bar.html.spt', "Greetings program!"),)
    assert_match(harness, '/foo/', 'foo/%bar.html.spt', wildcards={'bar': ''})

def test_missing_trailing_slash_matches_wild_leaf(harness):
    harness.fs.www.mk(('foo/%bar.html.spt', "Greetings program!"),)
    assert_match(harness, '/foo', 'foo/%bar.html.spt',
                 wildcards={'bar': ''}, canonical='/foo/')

def test_dont_confuse_files_for_dirs(harness):
    harness.fs.www.mk(('foo.html', 'Greetings, Program!'),)
    assert_missing(harness, '/foo.html/')
    assert_missing(harness, '/foo.html/bar')

def test_wildleaf_with_extention_doesnt_match_trailing_slash(harness):
    harness.fs.www.mk(
        ('%name/index.html.spt', ''),
        ('%name/%cheese.txt.spt', ''),
    )
    assert_missing(harness, '/chad/cheddar.txt/')

def test_wildleaf_without_extention_matches_trailing_slash(harness):
    harness.fs.www.mk(
        ('%name/index.html.spt', ''),
        ('%name/%cheese.spt', ''),
    )
    assert_match(harness, '/chad/cheddar.txt/', '%name/%cheese.spt',
                 wildcards={'name': 'chad', 'cheese': 'cheddar.txt/'})

def test_extra_slash_matches(harness):
    harness.fs.www.mk(('foo/bar.spt', ''))
    assert_match(harness, '/foo/bar/', 'foo/bar.spt', canonical='/foo/bar')

def test_resource_in_parent_directory_is_used_as_index(harness):
    harness.fs.www.mk(('foo/bar.spt', ''))
    harness.fs.www.mk(('foo/bar/baz.spt', ''))
    assert_match(harness, '/foo/bar/', 'foo/bar.spt')


# path part params
# ================

def test_path_part_with_params_works(harness):
    harness.fs.www.mk(('foo/index.html', "Greetings program!"),)
    assert_match(harness, '/foo;a=1/', '/foo/index.html')

def test_path_part_params_vpath(harness):
    harness.fs.www.mk(('%bar/index.html', "Greetings program!"),)
    assert_match(harness, '/foo;a=1;b=;a=2;b=3/', '/%bar/index.html', wildcards={'bar': 'foo'})

def test_path_part_params_static_file(harness):
    harness.fs.www.mk(('/foo/bar.html', "Greetings program!"),)
    assert_match(harness, '/foo/bar.html;a=1;b=;a=2;b=3', '/foo/bar.html')

def test_path_part_params_simplate(harness):
    harness.fs.www.mk(('/foo/bar.html.spt', NEGOTIATED_SIMPLATE),)
    assert_match(harness, '/foo/bar.html;a=1;b=;a=2;b=3', '/foo/bar.html.spt')

def test_path_part_params_negotiated_simplate(harness):
    harness.fs.www.mk(('/foo/bar.spt', NEGOTIATED_SIMPLATE),)
    assert_match(harness, '/foo/bar.txt;a=1;b=;a=2;b=3', '/foo/bar.spt', extension='txt')

def test_path_part_params_greedy_simplate(harness):
    harness.fs.www.mk(('/foo/%bar.spt', NEGOTIATED_SIMPLATE),)
    assert_match(harness, '/foo/baz/buz;a=1;b=;a=2;b=3/blam.html', '/foo/%bar.spt',
                 wildcards={'bar': 'baz/buz/blam.html'})


# mongs
# =====
# These surfaced when porting mongs from Aspen 0.8.

def test_virtual_path_parts_can_be_empty(harness):
    harness.fs.www.mk(('foo/%bar/index.html.spt', NEGOTIATED_SIMPLATE),)
    assert_wildcards(harness, '/foo//' , {'bar': ''})

def test_file_matches_in_face_of_dir(harness):
    harness.fs.www.mk( ('%page/index.html.spt', NEGOTIATED_SIMPLATE)
          , ('%value.txt.spt', NEGOTIATED_SIMPLATE)
           )
    assert_wildcards(harness, '/baz.txt', {'value': 'baz'})

def test_file_doesnt_match_if_extensions_dont_match(harness):
    harness.fs.www.mk(
        ('%page/index.html.spt', NEGOTIATED_SIMPLATE),
        ('%value.txt.spt', NEGOTIATED_SIMPLATE)
    )
    assert_wildcards(harness, '/baz.html', {'page': 'baz.html'})

def test_file_matches_extension(harness):
    harness.fs.www.mk( ('%value.json.spt', '[-----]\n[-----]\n{"Greetings,": "program!"}')
          , ('%value.txt.spt', NEGOTIATED_SIMPLATE)
           )
    assert_match(harness, '/baz.json', "%value.json.spt", wildcards={'value': 'baz'})

def test_file_matches_other_extension(harness):
    harness.fs.www.mk( ('%value.json.spt', '[-----]\n[-----]\n{"Greetings,": "program!"}')
          , ('%value.txt.spt', NEGOTIATED_SIMPLATE)
           )
    assert_match(harness, '/baz.txt', "%value.txt.spt", wildcards={'value': 'baz'})


def test_virtual_file_with_no_extension_works(harness):
    harness.fs.www.mk(('%value.spt', NEGOTIATED_SIMPLATE),)
    assert_match(harness, '/baz.txt', '%value.spt', wildcards={'value': 'baz.txt'})

def test_normal_file_with_no_extension_works(harness):
    harness.fs.www.mk(
        ('%value.spt', NEGOTIATED_SIMPLATE),
        ('value', '{"Greetings,": "program!"}')
    )
    assert_match(harness, '/baz.txt', '%value.spt', wildcards={'value': 'baz.txt'})

def test_file_with_no_extension_matches(harness):
    harness.fs.www.mk( ('%value.spt', NEGOTIATED_SIMPLATE)
          , ('value', '{"Greetings,": "program!"}')
           )
    assert_match(harness, '/baz', '%value.spt', wildcards={'value': 'baz'})

def test_dont_serve_hidden_files(harness):
    harness.fs.www.mk(('.secret_data', ''),)
    assert_missing(harness, '/.secret_data')

def test_do_serve_files_under_well_known_directory(harness):
    harness.fs.www.mk(('.well-known/security.txt', 'Lorem ipsum'),)
    assert_match(harness, '/.well-known/security.txt', '.well-known/security.txt')

def test_dont_serve_spt_file_source(harness):
    harness.fs.www.mk(('foo.html.spt', "Greetings, program!"),)
    assert_missing(harness, '/foo.html.spt')


# dispatch_result.extension
# =========================

def test_dispatcher_sets_extension(harness):
    harness.fs.www.mk(('foo.spt', "Greetings, program!"))
    assert_match(harness, '/foo.css', 'foo.spt', extension='css')

def test_extension_is_set_even_when_unknown(harness):
    harness.fs.www.mk(('foo.spt', "Greetings, program!"))
    assert_match(harness, '/foo.unknown-extension', 'foo.spt', extension='unknown-extension')

def test_extension_is_None_when_url_doesnt_contain_any(harness):
    harness.fs.www.mk(('foo.spt', "Greetings, program!"))
    assert_match(harness, '/foo', 'foo.spt', extension=None)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import codecs


# Before Python 3.5 the 'backslashreplace' error handler only worked when
# encoding, not when decoding. We want the 3.5+ behavior, and we want it in the
# global codecs registry as soon as possible.

def backslashreplace_errors(unicode_error):
    offender = unicode_error.object[unicode_error.start:unicode_error.end]
    if isinstance(offender, bytes):
        r = ''.join(r'\x{0:x}'.format(b if isinstance(b, int) else ord(b))
                    for b in offender)
    else:
        r = offender.encode('ascii', 'old-backslashreplace').decode('ascii')
    return (r, unicode_error.end)

def upgrade_backslashreplace():
    if codecs.lookup_error('backslashreplace') is not codecs.backslashreplace_errors:
        return  # don't upgrade if someone else already has
    codecs.register_error('old-backslashreplace', codecs.backslashreplace_errors)
    codecs.register_error('backslashreplace', backslashreplace_errors)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import sys
import pkg_resources

# Built-in renderers
BUILTIN_RENDERERS = [ 'stdlib_format'
                    , 'stdlib_percent'
                    , 'stdlib_template'
                    , 'json_dump'
                    , 'jsonp_dump'
                     ]

RENDERERS = BUILTIN_RENDERERS[:]

for entrypoint in pkg_resources.iter_entry_points(group='aspen.renderers'):
    RENDERERS.append(entrypoint.name)

RENDERERS.sort()

def factories(configuration):
    """return a dict of render factory names to the factories themselves"""
    renderer_factories = {}
    # import the built-in renderers
    for name in BUILTIN_RENDERERS:
        # Pre-populate renderers so we can report on ImportErrors early
        try:
            capture = {}
            python_syntax = 'from aspen.simplates.renderers.%s import Factory'
            exec(python_syntax % name, capture)
            make_renderer = capture['Factory'](configuration)
        except ImportError as err:
            make_renderer = err
            err.info = sys.exc_info()
        renderer_factories[name] = make_renderer

    # import renderers provided by other packages
    for entrypoint in pkg_resources.iter_entry_points(group='aspen.renderers'):
        render_module = entrypoint.load()
        renderer_factories[entrypoint.name] = render_module.Factory(configuration)
    return renderer_factories


# abstract bases
# ==============
# The base is actually functional. It's a pass-through.

class Renderer(object):
    """The base class of renderers.
    """

    def __init__(self, factory, filepath, raw, media_type, offset):
        """Takes a Factory, three bytestrings, and an int.
        """
        self._filepath = filepath
        self._factory = factory
        self._changes_reload = factory._changes_reload
        self.meta = self._factory.meta
        self.raw = raw
        self.media_type = media_type
        self.offset = offset
        self.padded = ('\n' * offset) + self.raw
        self.compiled = self.compile(self._filepath, self.padded)

    def __call__(self, context):
        if self._changes_reload:
            self.meta = self._factory._update_meta()
            self.compiled = self.compile(self._filepath, self.padded)
        r = self.render_content(context)
        if r[:self.offset] == self.padded[:self.offset]:
            # The padding is still there, strip it
            return r[self.offset:]
        return r

    def compile(self, filepath, padded):
        """Override.

        Whatever you return from this will be set on self.compiled the first
        time the renderer is called. If changes_reload is True then this will
        be called every time the renderer is called. You can then use
        self.compiled in your render_content method as needed.

        """
        return padded

    def render_content(self, context):
        """Override. Context is a dict.

        You can use these attributes::

            self.raw        the raw bytes of the content page
            self.compiled   the result of self.compile (generally a template in
                             compiled object form)
            self.meta       the result of Factory.compile_meta
            self.media_type the media type of the page
            self.offset     the line number at which the page starts

        """
        return self.raw  # pass-through


class Factory(object):
    """The base class of renderer factories.
    """

    Renderer = Renderer

    def __init__(self, configuration):
        self._configuration = configuration
        self._changes_reload = configuration.changes_reload
        self.meta = self.compile_meta(configuration)

    def __call__(self, filepath, raw, media_type, offset):
        """Given three bytestrings and an int, return a callable.
        """
        self._update_meta()
        return self.Renderer(self, filepath, raw, media_type, offset)

    def _update_meta(self):
        if self._changes_reload:
            self.meta = self.compile_meta(self._configuration)
        return self.meta  # used in our child, Renderer

    def compile_meta(self, configuration):
        """Takes a configuration object. Override as needed.

        Whatever you return from this will be set on self.meta the first time
        the factory is called, or every time if changes_reload is True. You can
        then use self.meta in your Renderer class as needed.

        """
        return None



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import sys

from pytest import raises, mark

from aspen.request_processor import ConfigurationError, RequestProcessor
from aspen.testing import chdir


def test_defaults_to_defaults(harness):
    os.chdir(harness.fs.www.root)
    rp = RequestProcessor()
    actual = ( rp.project_root
             , rp.www_root

             , rp.changes_reload
             , rp.charset_static
             , rp.encode_output_as
             , rp.indices
             , rp.media_type_default
             , rp.media_type_json
             , rp.renderer_default
              )
    expected = ( None, os.getcwd(), False, None, 'UTF-8'
               , ['index.html', 'index.json', 'index', 'index.html.spt', 'index.json.spt', 'index.spt']
               , 'text/plain', 'application/json', 'stdlib_percent'
                )
    assert actual == expected

@mark.skipif(sys.platform == 'win32',
             reason="Windows file locking makes this fail")
def test_ConfigurationError_raised_if_no_cwd(harness):
    FSFIX = harness.fs.project.resolve('')
    os.chdir(FSFIX)
    os.rmdir(FSFIX)
    raises(ConfigurationError, RequestProcessor)

@mark.skipif(sys.platform == 'win32',
             reason="Windows file locking makes this fail")
def test_ConfigurationError_NOT_raised_if_no_cwd_but_do_have__www_root(harness):
    project_root = harness.fs.project.root
    www_root = harness.fs.www.root
    with chdir(project_root):
        os.rmdir(project_root)
        rp = RequestProcessor(www_root=www_root)
        assert rp.www_root == www_root

def test_processor_sees_root_option(harness):
    rp = RequestProcessor(www_root=harness.fs.project.resolve(''))
    expected = harness.fs.project.root
    actual = rp.www_root
    assert actual == expected

def test_user_can_set_renderer_default(harness):
    SIMPLATE = """
[----]
name="program"
[----]
Greetings, {name}!
    """
    harness.fs.www.mk(('index.html.spt', SIMPLATE),)
    harness.hydrate_request_processor(renderer_default="stdlib_format")
    actual = harness.simple(filepath=None, uripath='/', want='output.text')
    assert actual == 'Greetings, program!\n'



from .utils import auto_repr


@auto_repr
class Output(object):
    """The result of rendering a resource.
    """

    __slots__ = ('body', 'media_type', 'charset')

    def __init__(self, body=None, media_type=None, charset=None):
        self.body = body
        self.media_type = media_type
        self.charset = charset

    @property
    def text(self):
        return self.body.decode(self.charset) if self.charset else None



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import sys
from warnings import catch_warnings

from aspen.exceptions import AttemptedBreakout, PossibleBreakout
from aspen.http.resource import open_resource
from aspen.simplates.pagination import split
import pytest
from pytest import raises


# Test outputs

def test_barely_working(harness):
    output = harness.simple('Greetings, program!', 'index.html')
    assert output.media_type == 'text/html'

def test_charset_static_barely_working(harness):
    output = harness.simple( ('Greetings, program!', 'utf16')
                           , 'index.html'
                           , request_processor_configuration={'charset_static': 'utf16'}
                            )
    assert output.media_type == 'text/html'
    assert output.charset == 'utf16'

def test_charset_static_checks_encoding(harness):
    output = harness.simple( ('Greetings, program!', 'utf16')
                           , 'index.html'
                           , request_processor_configuration={'charset_static': 'utf8'}
                            )
    assert output.media_type == 'text/html'
    assert output.charset is None

def test_charset_static_None(harness):
    output = harness.simple( 'Greetings, program!'
                           , 'index.html'
                           , request_processor_configuration={'charset_static': None}
                            )
    assert output.media_type == 'text/html'
    assert output.charset is None

def test_encode_output_as_barely_working(harness):
    output = harness.simple( '[---]\n[---]\nGreetings, program!'
                           , 'index.html.spt'
                           , request_processor_configuration={'encode_output_as': 'ascii'}
                            )
    assert output.media_type == 'text/html'
    assert output.charset == 'ascii'

def test_resource_pages_work(harness):
    actual = harness.simple("[---]\nfoo = 'bar'\n[--------]\nGreetings, %(foo)s!").text
    assert actual == "Greetings, bar!"

def test_resource_dunder_all_limits_vars(harness):
    actual = raises( KeyError
                   , harness.simple
                   , "[---]\nfoo = 'bar'\n"
                     "__all__ = []\n"
                     "[---------]\n"
                     "Greetings, %(foo)s!"
                    ).value
    # in production, KeyError is turned into a 500 by an outer wrapper
    assert type(actual) == KeyError

def test_path_part_params_are_available(harness):
    output = harness.simple("""
        [---]
        if 'b' in path.parts[0].params:
            a = path.parts[0].params['a']
        [---]
        %(a)s
    """, '/foo/index.html.spt', '/foo;a=1;b;a=3/')
    assert output.text == "3\n"

def test_resources_dont_leak_whitespace(harness):
    """This aims to resolve https://github.com/whit537/aspen/issues/8.
    """
    actual = harness.simple("""
        [--------------]
        foo = [1,2,3,4]
        [--------------]
        %(foo)r""").text
    assert actual == "[1, 2, 3, 4]"

def test_negotiated_resource_doesnt_break(harness):
    expected = "Greetings, bar!\n"
    actual = harness.simple("""
        [-----------]
        foo = 'bar'
        [-----------] text/plain
        Greetings, %(foo)s!
        [-----------] text/html
        <h1>Greetings, %(foo)s!</h1>
        """
        , filepath='index.spt').text
    assert actual == expected

def test_request_processor_is_in_context(harness):
    output = harness.simple("""
        assert request_processor.__class__.__name__ == 'RequestProcessor', request_processor
        [--------]
        [--------]
        It worked.""")
    assert output.text == 'It worked.'

def test_unknown_mimetype_yields_default_mimetype(harness):
    output = harness.simple( 'Greetings, program!'
                             , filepath='foo.flugbaggity'
                              )
    assert output.media_type == 'text/plain'

def test_templating_without_script_works(harness):
    output = harness.simple('[-----]\n[-----] via stdlib_format\n{path.raw}')
    assert output.text == '/'


# Test offset calculation

def check_offsets(raw, offsets):
    actual = [page.offset for page in split(raw)]
    assert actual == offsets

def test_offset_calculation_basic(harness):
    check_offsets('\n\n\n[---]\n\n', [0, 4])

def test_offset_calculation_for_empty_file(harness):
    check_offsets('', [0])

def test_offset_calculation_advanced(harness):
    raw = (
        '\n\n\n[---]\n'
        'cheese\n[---]\n'
        '\n\n\n\n\n\n[---]\n'
        'Monkey\nHead\n') #Be careful: this is implicit concation, not a tuple
    check_offsets(raw, [0, 4, 6, 13])


# Test the `open_resource` function

def test_open_resource_opens_file_inside_www_root(harness):
    harness.fs.www.mk(('index.html', 'foo'))
    fspath = harness.fs.www.resolve('index.html')
    with open_resource(harness.request_processor, fspath) as f:
        actual = f.read()
        assert actual == b'foo'

def test_open_resource_opens_file_inside_project_root(harness):
    harness.fs.project.mk(('error.spt', 'bar'))
    spt_path = harness.fs.project.resolve('error.spt')
    with open_resource(harness.request_processor, spt_path) as f:
        actual = f.read()
        assert actual == b'bar'

# `realpath` doesn't work on Windows in Python < 3.8: https://bugs.python.org/issue9949
@pytest.mark.xfail('os.path.realpath is os.path.abspath')
def test_open_resource_raises_exception_when_file_is_outside(harness):
    # Create a symlink, if possible.
    executable = os.path.realpath(sys.executable)
    fspath = harness.fs.www.resolve('python')
    try:
        os.symlink(executable, fspath)
    except NotImplementedError:
        return
    # Initialize the request processor and check the warnings.
    with catch_warnings(record=True) as messages:
        harness.hydrate_request_processor()
        assert messages
        for m in messages:
            warning = m.message
            assert isinstance(warning, PossibleBreakout)
            assert warning.sym_path == fspath
            assert warning.real_path == executable
    # Attempt to open the resource.
    with raises(AttemptedBreakout):
        open_resource(harness.request_processor, fspath)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from . import Renderer, Factory
from .. import json_

class Renderer(Renderer):

    def render_content(self, context):
        output = context['output']
        if not output.media_type:
            output.media_type = context['request_processor'].media_type_json
        r = json_.dumps(eval(self.compiled, globals(), context))
        if isinstance(r, bytes):
            r = r.decode('ascii')
        return r


class Factory(Factory):
    Renderer = Renderer

